(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)
(* Command Line version of polymorphic coercion compiler *)

open Lambd
open Typaux
open Typing
open Pars
open Lexer
open Sys
open Util

let print_only = ref false

let version_string = "Coco: Compiler for Monadic Coercions\nVersion 0.2"
let print_version () = 
  print_string version_string; print_newline()

let regurgitate_input (s:signature) (g:typ_env) (e:stritem) =
  if not (empty g) then print_string
    ("extern{\n  " ^ String.concat ",\n  "
       (List.map (fun (x,s) -> x ^ " : " ^ string_of_scheme s) g)
    ^ "\n}\n");
  if not (empty s) then print_string
    ("coercion{\n  " ^ string_of_signature "\n  " s ^ "\n}\n");
  Format.open_box 1;
  print_stritem (s,[]) e;
  Format.print_newline()


let typing_str s g e =
  Format.open_box 1; print_stritem (s,[]) e; Format.print_newline();
  try (
    match Typing.typing_str s g e with
    | str,_,None -> 
      print_stritem (s,[]) str; Format.print_newline(); 
      () (*Typaux.print_ml ~output_file:!ml_output str s
*)
    | str,_,Some (e',ot,cs) -> 
        let t = norm_ortyp ot in
        if not !quiet then begin
          Format.print_string " ~~> "; Format.print_newline();
	  print_stritem (s,[]) (StrCons(str,StrTerm e')); Format.print_newline(); 
          print_string ("TYPE: " ^ string_of_ortyp ot ^ "\n");
          Format.print_string " ~~> "; Format.print_newline();
        end;
        let inimlfile = (Filename.chop_extension !infile)^"_pre.ml" in
(*        if not !Util.quiet 
        then Typaux.print_ml ~output_file:inimlfile (StrCons(str,StrTerm e')) s;
*)        Solver.strict := !Util.strict;
        let unsolved,theta = 
	  if not !Util.solve then cs,[]
	  else Solver.solve ~where:"toplevel" s cs (Typing.fmvor t) in
        Format.open_box 1; 
        Format.print_string " ~~> "; Format.print_newline();
 
        let abs_e = e' (*abstract_constraints e' unsolved s*) in
	let e' = List.fold_right rename_binds(*subst_term*) theta abs_e in
        Solver.strict := true;
        let pretty_e = StrCons(str,StrTerm (simplify s g e')) in
	print_stritem (s,[]) pretty_e; Format.print_newline(); 
(*        Typaux.print_ml ~output_file:!ml_output pretty_e s;
*)        if not (empty unsolved) then begin
              let fmvars = (*fmvor t*) (union_all (List.map fmvc unsolved)) in
              let forall = match fmvars with 
                | [] -> "" 
                | _ -> if !Util.latex then "\\forall "^(String.concat "," (List.map mvar2tex fmvars))^". "
                       else "forall "^(String.concat "," fmvars)^". " in
              print_string ("Unsolved constraints:");
              print_string (" " ^ forall ^ string_of_constrs (if !Util.latex then ", " else " ") unsolved ^ "\n");
              if !Solver.dot then
              Solver.dump_graph (Solver.graph_from_constraints unsolved) (!infile^"-unsolved.dot")
          else ()
        end;
        print_string ("TYPE: " ^ string_of_ortyp t ^ "\n");
	Format.close_box()
)       with e -> (Format.print_string "FAILED"; Format.print_newline(); Format.close_box(); raise e)
;;


let arg_specs =
  [
   ("-print", Arg.Set print_only, "Pretty-print input only");
   ("-nodebug", Arg.Clear Util.debug, "Instantiate types");
   ("-quiet", Arg.Set quiet, "Don't print version preamble and debugging info");
   ("-dot", Arg.Set Solver.dot, "Write .dot file with initial and final constraints");
   ("-strict", Arg.Set Util.strict, "Refuse unbounded monad variables");
   ("-letgen",Arg.Set Util.letgen, "Generalize all let-bound definitions");
   ("-lift",Arg.Set Typing.lift, "Lift pure computations to Bot monad");
   ("-nosolve", Arg.Clear Util.solve, "Do not solve any generated constraints");
   ("-notindexed", Arg.Clear Util.indexed, "Do not support type indexed monads");
   ("-mloutput",Arg.Set_string ml_output, "Write the result of OCaml conversion to file");
   ("-latex",Arg.Set Util.latex, "Output the constraints in latex")
 ];;

let _ =
    Camlp4_config.quotations := false;
    let anon_arg s = 
      if Filename.check_suffix s ".coco" then
        if !infile = "" then infile := s 
        else print_endline ("Warning: ignoring "^s^"; .coco already given!")
      else 
      if Filename.check_suffix s ".ml" then
        if !ml_input = "-" then ml_input := s 
        else print_endline ("Warning: ignoring "^s^"; .coco already given!")
      else 
      if Filename.check_suffix s ".mli" 
      then mli_input := s :: !mli_input
      else print_endline ("Warning: ignoring "^s^"; unknown extension.") in
    Arg.parse arg_specs anon_arg version_string;
    mli_input := List.rev !mli_input;
    if !infile = "" 
    then failwith "A .coco file with monad declarations should be provided";
    if !ml_output = "" then begin
      let dir,base = split_filename (if !ml_input="-" then !infile else !ml_input) in    
      ml_output := Filename.concat dir (base^"_result.ml")
    end;

    if !Util.strict then Util.letgen := true;
    if not !quiet then print_version (); 

    try (
        let chan_ml = open_in !infile in
	let lexbuf = Lexing.from_channel chan_ml  in
        (* Parse the monadic declarations *)
        let ordering,s_input = Pars.start Lexer.token lexbuf in
        (*Solver.init_ordering ordering; *)
        close_in chan_ml;
        (* Parse the external .mli interface *)
        let g_input,s_new = List.fold_left (fun (sg2,ts2) fmli -> 
            let g_mli = Typaux.parse_ocamli fmli in
            let sg1,ts1 = (Typaux.sig_of_ocaml ts2 g_mli) in
            sg1@sg2, ts1
          ) ([],s_input) !mli_input
        in
        (* Parse the .ml code to be converted *)
        let e = 
              if !ml_input = "" then begin
                print_endline "Program to convert is expected on stdin:";
                Typaux.parse_ocaml stdin s_new end
              else 
                let chan = open_in !ml_input in
                let res = Typaux.parse_ocaml chan s_new in
                close_in chan;
                res in
        (* Generate the .mli interface for the monadic declarations *) 
	let dir,base = split_filename !ml_output in
	let mlifile = Filename.concat dir ("lib"^base^".mli") in
(*        print_mli mlifile [] s_input;
*)      let s = (Monad(bottom,botBind,botUnit,[])) :: s_new in
        let g = let va = TVar("a",ref None) in
          ("id", Scheme(["a"],[],TFun(va,(MBase(bottom,[]),va)))) :: g_input in
        Solver.check_wf_morphisms s 0; 
        if !print_only then
          regurgitate_input s g e
	else
          typing_str s g e
    ) with 
        Parsing.Parse_error -> 
	  (print_endline ("Oops! Error in lexing/parsing "^ !infile);
	  exit 1)
        | exc ->
            Format.eprintf "error: %a@." Camlp4.ErrorHandler.print exc;
            exit 2 
