(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)
open Util

type ident = string
type monad =
  | MVar of ident * (monad option ref) (* monad constructor var *)
  | MBase of ident * typ list (* monad costructor *)
and typ =
    TBase of ident        (* generic base type *)
  | TVar of ident * (typ option ref)
  | TFun of typ * mtyp
  | TTup of typ list
  | TApp of typ * typ list
  | TMorph of mtyp * mtyp
and mtyp = monad * typ

(* bind and unit for the bottom monad *)
let bottom = "bot"
let mBot = MBase(bottom,[])
let botBind = "rApply"
let botUnit = "id"

type label = ident
type tvar = ident
type var = ident
type constr = (label * monad * monad) list * monad

type scheme = Scheme of tvar list * constr list * typ
type ortyp = Pure of typ | Monadic of mtyp

type typlet = TLet of scheme | TBind of mtyp

type term =
    Base of ident * string  (* generic base values; ident is the type name *)
(*  | Cons of ident * term list *)
  | Var of ident
  | Pair of term * term
  | Tuple of term list
  | Index of bool * term  (* true==fst, false==snd *)
  | Lam of ident * typ option * term (* annotation is for the target language *) 
  | Rec of ident * term
  | App of term * term
  | Let of ident * term * term
  | Seq of term * term
  | Cond of term * term * term
  | Annot of term * typlet (* only used for printing intermediate types *)
  | Susp of (term -> term) (* only used during evaluation *) 
  | Bind of term * (monad * monad * monad) * term * term (* target language: monadic bind *)
  | Unit of monad * term (* target language: monadic unit *)
  | EvApp of ident * monad list * term list (* target language: application of evidence to a var *)
  | EvAbs of ident list * (ident * typ option) list * term (* target language: abstraction of evidence *)

(* A subgrammar for toplevel OCaml constructs *)
type stritem = 
  | StrEmpty
  | StrTerm of term
  | StrLet of bool * var * term
  | StrCons of stritem * stritem

(************************************************************************)
exception Type_error of string
let gen_lab =
  let x = ref 0 in
  (function () ->
    incr x; "l"^(string_of_int !x))

let freshvar =
  let x = ref 0 in
  (function () ->
    incr x; "v"^(string_of_int !x))

let freshtvar : unit -> typ =
  let x = ref 0 in
  (function () ->
    incr x; TVar ("X"^(string_of_int !x),ref None))

let freshmvar : unit -> var * monad =
  let x = ref 0 in
  (function () ->
    incr x; 
    let name = "m"^(string_of_int !x) in 
    name, MVar (name,ref None))

let is_tmpvar v = 
  try Scanf.sscanf v "v%d" (fun _ -> true)
  with Scanf.Scan_failure _ -> false

let identity = let f = freshvar () in Lam(f,None,Var f)

(* Syntax builders *)

let rec apply f = function 
    [] -> f
  | e::rest -> apply (App(f,e)) rest

let rec lam bs e =
  match bs with
    [] -> e
  | (x,t)::rest -> Lam(x,t,lam rest e)

let rec tuple = function
    [] -> failwith "tuple: size-0 list"
  | [e] -> e
  | e::rest -> Pair(e,tuple rest)


let rec mflatten m =
  match m with
    MVar (_,r) ->
      (match !r with
        None -> m
      | Some m' -> mflatten m')
  | _ -> m

(* free variables of a term *)
let rec fv_term = function
    Base _ | Susp _ -> []
(*  | Cons (v,l) -> union_all (List.map fv_term l ) *)
  | Var y -> [y] 
  | EvApp (y,_,l) -> union_all (List.map fv_term l) 
  | Tuple el -> union_all (List.map fv_term el) 
  | Index (b,e) ->
      fv_term e
  | Lam (x1,_,e) ->
      setminus [x1] (fv_term e)
  | Rec (x1,e) -> setminus [x1] (fv_term e)
  | App (e1,e2)
  | Let (_,e1,e2)
  | Pair (e1,e2) 
  | Seq (e1,e2) -> 
      let s1 = fv_term e1 in
      let s2 = fv_term e2 in
      union s1 s2
  | Cond (e1,e2,e3) ->
      let s1 = fv_term e1 in
      let s2 = fv_term e2 in
      let s3 = fv_term e3 in
      union s1 (union s2 s3)
  | Annot(e,t1) -> 
      fv_term e
  | Bind (v,_,e1,e2) -> (fv_term v)@(union (fv_term e1) (fv_term e2))
  | Unit (_,e) | EvAbs (_,_,e) -> fv_term e


let rec fmvmonad = function 
    MVar (m,r) -> [m] 
  | _ -> []
let rec fmv = function
    TBase _ -> []
  | TFun (t1,t2) ->  
      let fmv1 = fmv t1 in
      let fmv2 = fmvm t2 in
      union fmv1 fmv2
  | TTup (tl) -> union_all (List.map fmv tl)
  | TVar (tv,r) -> []
  | TApp (t, tl) -> 
      union_all (List.map fmv (t::tl))
  | TMorph (m1,m2) -> union (fmvm m1) (fmvm m2)
and fmvm (m,t) = union (fmvmonad m) (fmv t)
let fmvor = function
  | Pure t -> fmv t
  | Monadic t -> fmvm t
let fmvc (ml,m) = (union_all 
    ((fmvmonad (mflatten m)) :: 
    (List.map (fun (_,m1,m2) -> union 
       (fmvmonad (mflatten m1)) (fmvmonad (mflatten m2))) ml))) 

let rec ftv = function
    TBase _ -> []
  | TMorph (t1,t2) -> union (ftvm t1) (ftvm t2)
  | TFun (t1,t2) ->  
      let ftv1 = ftv t1 in
      let ftv2 = ftvm t2 in
      union ftv1 ftv2
  | TTup (tl) -> union_all (List.map ftv tl)
  | TVar (tv,r) ->
      (match !r with
        None -> [tv]
         | Some t -> ftv t)
  | TApp (t, tl) -> 
      union_all (List.map ftv (t::tl))
and ftvm (m,t) = union (ftvmonad m) (ftv t)
and ftvmonad = function 
    MVar (m,r) -> (match !r with None -> [m] | Some l ->  ftvmonad l) 
  | MBase (m,tl) -> union_all (List.map ftv tl)

let ftvc (ml,m2) = union_all ((ftvmonad m2) 
  :: List.map (fun (l,m1,m2)-> union (ftvmonad m1) (ftvmonad m2)) ml) 
let ftvcs cs = union_all (List.map ftvc cs)


let ftvs (Scheme (tvs,cs,typ)) =
  setminus (union (ftv typ) (ftvcs cs)) tvs

let ftvg g = union_all (List.map (fun (x,s) -> ftvs s) g)

let vofc (lab,_,_) = lab
let csof (Scheme (_,cs,_)) = cs

(************************************************************************)
(* Ugly printing *)

let paren s = "("^s^")"
let mvar2tex s = "\\mvar_{"^(String.sub s 1 (String.length s -1))^"}"

let rec string_of_monad = function
  | MVar (s,r) ->
    (match !r with 
        None -> if not !Util.latex then s else mvar2tex s
      | Some m -> string_of_monad m)
  | MBase (s,[]) -> if s = bottom && !Util.latex then "\\mId" else s
  | MBase (s,tl) -> s^" "^(String.concat " "(List.map string_of_typ tl))
and string_of_typ = function
    TBase s -> s
  | TFun (t1,t2) -> 
    let arrow = if !Util.latex then "\\rightarrow" else "->" in
    paren ((string_of_typ t1)^arrow^(string_of_mtyp t2))
  | TTup (tl) -> paren ((String.concat "*" (List.map string_of_typ tl)) )
  | TVar (tv,r) ->
      (match !r with
           None -> tv
         | Some t -> string_of_typ t)
  | TApp (t1, []) -> (string_of_typ t1) 
  | TApp (t1, tl) -> "((" ^ (String.concat "," (List.map string_of_typ tl)) ^ ")" ^ (string_of_typ t1) ^")" 
  | TMorph (m1,m2) -> (string_of_mtyp m1) ^ "<<" ^ (string_of_mtyp m2)
and string_of_mtyp (m,t) = 
  (string_of_monad m) ^ (if !Util.latex then "\\;" else " ") ^ (string_of_typ t)

let string_of_ortyp = function
  | Pure t -> string_of_typ t
  | Monadic m -> string_of_mtyp m

let string_of_topt = function
    Some t -> ":"^(string_of_typ t)^" "
  | None -> ""

let string_of_constr (cs,target) = 
  let cres = if not !Util.latex then " << "^string_of_monad target 
     else string_of_monad target ^"}" in
  let f = fun (l,m1,m2) -> if not !Util.latex 
    then l^": ("^string_of_monad m1^"+"^string_of_monad m2^")"^cres
    else "\\bind{("^string_of_monad m1^","^string_of_monad m2^")}{"^cres in
  let css = List.map f cs in
  if not !Util.latex then
    "("^String.concat ") (" css^")"
  else
    String.concat ", " css


let string_of_constrs sep cs =
  String.concat ("" ^ sep) (List.map string_of_constr cs)

let string_of_scheme (Scheme (tvs,cs,typ)) =
  let not_in_sig c = not (Util.empty (fmvc c)) in
  let cs = List.filter not_in_sig cs in
  let tvs = intersect tvs (ftvcs cs) in 
  let tvs_str = 
    match tvs with [] -> "" | _  -> 
      if !Util.latex then "\\forall "^(String.concat "," (List.map mvar2tex tvs))^". "
                       else "forall "^(String.concat "," tvs)^". " in
  let cs_str = 
    match cs with [] -> "" | _  -> "{"^
     (string_of_constrs (if !Util.latex then ". \ lam" else ",") cs)^"}" in
  tvs_str ^ cs_str ^ (string_of_typ typ)


let rec string_of_term = function
    Base (t,v) -> 
      if (t = "int" || t = "bool") then v
      else if t = "unit" then "()"
      else if (t = "string") then ("\"" ^ v ^ "\"")
      else "{"^t^":"^v^"}"
(*  | Cons (v,l) -> v ^ (match l with [] -> "" 
	| [x::l2] -> "("^string_of_term x ^ (List.map string_of_term l2)^ ")") *)
  | Var x -> x
  | Tuple el -> "("^(String.concat "," (List.map string_of_term el))^")"
  | Pair (e1,e2) -> 
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      "<"^s1^","^s2^">"
  | Index (b,e) ->
      let s = string_of_term e in
      if b then "fst "^s else "snd "^s
  | Lam (x,t,e) ->
      let s = string_of_term e in
      let st = match t with None -> "" | Some x -> ":"^(string_of_typ x) in
      "fun "^x^st^"->"^s
  | Rec (x,e) ->
      let s = string_of_term e in
      "rec "^x^"."^s
  | App (e1,e2) ->
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      "("^s1^" "^s2^")"
  | Let (x,e1,e2) ->
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      "let "^x^" = "^s1^" in "^s2^" end"
  | Seq (e1,e2) -> 
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      s1^";\n"^s2
  | Cond (e1,e2,e3) ->
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      let s3 = string_of_term e3 in
      "if "^s1^" then "^s2^" else "^s3
  | Annot(e,t) -> 
      let s1 = string_of_term e in
      let s2 = match t with 
          TLet t -> string_of_scheme t 
        | TBind t -> string_of_mtyp t in
        s1 ^  ":" ^ s2
  | Susp k -> "<susp>"
  | Bind (v,(m1,m2,m3),e1,e2) -> 
     let ms = (string_of_monad m1^"_"^string_of_monad m2^"_"^string_of_monad m3) in
      let s1 = string_of_term e1 in
      let s2 = string_of_term e2 in
      "bind_"^ms^" "^s1^" "^s2
  | Unit (m,e) -> 
     let s = string_of_term e in
     let ms = string_of_monad m in
     "unit_"^ms^" "^s
  | EvApp (v,xl,l) ->
      v^" "^(String.concat " " (List.map string_of_monad xl))^" "^(String.concat " " (List.map string_of_term l)) 
  | EvAbs (x,l,e) ->
      let s = string_of_term e in
      "fun "^(String.concat " " (x@(List.map fst l)))^" -> "^s

let rec ecr t =
  match t with
    TVar (_,r) ->
      (match !r with
        None -> t
      | Some t -> ecr t)
  | _ -> t

(* - Substitute variable x for term t in expr 
 * - Simplify binds/units of the Bot monad; 
 * TODO: separate the simplification *)
let rec subst_term (x,t) expr = match expr with
    Base _ | Susp _ -> expr
  | Var y -> if x = y then t else Var y
  | Pair (e1,e2) -> 
      let s1 = subst_term (x,t) e1 in
      let s2 = subst_term (x,t) e2 in
      Pair(s1,s2)
  | Tuple (el) -> Tuple (List.map (subst_term (x,t)) el)
  | Index (b,e) ->
      let s = subst_term (x,t) e in
      Index(b,s)
  | Lam (x1,ty,e) ->
      let s = if x=x1 then e else subst_term (x,t) e in
      Lam(x1,ty,s)
  | Rec (x1,e) ->
      let s = if x=x1 then e else subst_term (x,t) e in
      Rec(x1,s)
  | App (e1,e2) ->
      let s1 = subst_term (x,t) e1 in
      let s2 = subst_term (x,t) e2 in
      (match s1 with 
        (*| Var x when x=identity -> s2 *) 
        (*| App (Var x, arg) when x=identity -> App(s2,arg) *)
	| Lam (x,_,App(Var fx,arg)) when (x=fx && not (List.mem x (fv_term arg))) ->
	    App (s2,arg)  
        | _ -> App(s1,s2))

  | Let (y,e1,e2) ->
      let s1 = subst_term (x,t) e1 in
      let s2 = if x=y then e2 else subst_term (x,t) e2 in
      Let(y,s1,s2)
  | Seq (e1,e2) ->
      let s1 = subst_term (x,t) e1 in
      let s2 = subst_term (x,t) e2 in
      Seq(s1,s2)
  | Cond (e1,e2,e3) ->
      let s1 = subst_term (x,t) e1 in
      let s2 = subst_term (x,t) e2 in
      let s3 = subst_term (x,t) e3 in
      Cond(s1,s2,s3)
  | Annot(e,t1) -> 
      let s1 = subst_term (x,t) e in
      Annot(s1,t1)
  | Bind (v,((m1,m2,m3) as m),e1,e2) -> 
      let s1 = subst_term (x,t) e1 in
      let s2 = subst_term (x,t) e2 in
      (match mflatten m1,mflatten m2,mflatten m3,s2 with 
      | MBase (ms1,[]),MBase (ms2,[]),MBase(ms3,[]),Lam(y,_,e2') 
         when ms1=bottom && ms2=bottom && ms3=bottom-> 
         if is_tmpvar y then subst_term (y,s1) e2' else Let(y,s1,e2') 
      | MBase (ms1,[]),MBase (ms2,[]),m3',Lam(y,_,e2') 
         when ms1=bottom && ms2=bottom -> print_endline ("UNIT "^string_of_monad m3');
         if is_tmpvar y then Unit(m3',subst_term (y,s1) e2') else Unit(m3',App(s2,s1)) 
      | MBase (ms1,[]),m2',m3',Lam(y,_,e2') 
         when ms1=bottom && string_of_monad m2' = string_of_monad m3'-> 
         if is_tmpvar y then subst_term (y,s1) e2' else Let(y,s1,e2') 
      | m1',MBase (ms2,[]),m3',Lam(y,_,Var y') 
         when y=y' && ms2=bottom && string_of_monad m1' = string_of_monad m3'-> 
         s1 
      | _ -> Bind(v,m,s1,s2))
  | Unit(m,e) -> 
      let s1 = subst_term (x,t) e in
      (match mflatten m with MBase (s,[]) when s=bottom -> s1 | _ -> Unit(m,s1))
  | EvApp (v,m,l) -> if x=v then t else EvApp (v,m,List.map (subst_term (x,t)) l)
  | EvAbs (m,l,e) -> if List.mem x (List.map fst l) then expr else 
      let s1 = subst_term (x,t) e in
      EvAbs (m,l,s1)

let rec rename_binds (x,v) = function
  | Bind(Var z,(m1,m2,m3),e1,e2) ->  
      let s1 = rename_binds (x,v) e1 in
      let s2 = rename_binds (x,v) e2 in
      (match mflatten m1,mflatten m2,mflatten m3,s2 with 
      (*| MBase (ms1,[]),MBase (ms2,[]),MBase(ms3,[]),Lam(y,_,e2') 
         when ms1=bottom && ms2=bottom && ms3=bottom-> 
         if is_tmpvar y then subst_term (y,s1) e2' else Let(y,s1,e2') 
      | MBase (ms1,[]),MBase (ms2,[]),m3',Lam(y,_,e2') 
         when ms1=bottom && ms2=bottom ->
         if is_tmpvar y then Unit(m3',subst_term (y,s1) e2') else Unit(m3',App(s2,s1)) 
      | MBase (ms1,[]),m2',m3',Lam(y,_,e2') 
         when ms1=bottom && string_of_monad m2' = string_of_monad m3'-> 
         if is_tmpvar y then subst_term (y,s1) e2' else Let(y,s1,e2') 
      | m1',MBase (ms2,[]),m3',Lam(y,_,Var y') 
         when y=y' && ms2=bottom && string_of_monad m1' = string_of_monad m3'-> 
         s1 *)
      | m1',m2',m3',_ -> 
	if x=z then 
	  (if v = (Var botBind) then App(s2,s1) 
	  else Bind(v,(m1',m2',m3'),s1,s2))
	else Bind(Var z,(m1',m2',m3'),s1,s2))
  | Unit (m,e1) -> 
      let ne1 = rename_binds (x,v) e1 in
      Unit(m,ne1)
  | App (e1,e2) -> 
      let ne1 = rename_binds (x,v) e1 in
      let ne2 = rename_binds (x,v) e2 in
      App(ne1,ne2)
  | Cond (e1,e2,e3) -> 
      let ne1 = rename_binds (x,v) e1 in
      let ne2 = rename_binds (x,v) e2 in
      let ne3 = rename_binds (x,v) e3 in
      Cond(ne1,ne2,ne3)
  | Let (y,e2,e3) -> 
      let ne2 = rename_binds (x,v) e2 in
      let ne3 = rename_binds (x,v) e3 in
      Let (y,ne2,ne3)
  | Seq (e2,e3) -> 
      let ne2 = rename_binds (x,v) e2 in
      let ne3 = rename_binds (x,v) e3 in
      Seq (ne2,ne3)
  | Annot(e1,t) ->
      let ne1 = rename_binds (x,v) e1 in
      Annot(ne1,t)
  | Lam(y,t,e1) -> 
      let ne1 = rename_binds (x,v) e1 in
      Lam(y,t,ne1)
  | Rec(y,e1) -> 
      let ne1 = rename_binds (x,v) e1 in
      Rec(y,ne1)
  | Tuple(el) -> Tuple (List.map (rename_binds (x,v)) el)
  | Pair (e1,e2) -> 
      let ne1 = rename_binds (x,v) e1 in
      let ne2 = rename_binds (x,v) e2 in
      Pair (ne1,ne2)
  | e -> e

(* ref-based unification of monads and types*)
let rec unify_monad m1 m2 = let s1 = string_of_monad m1 in
            let s2 = string_of_monad m2 in

  match m1,m2 with
    MVar (m,r),_ -> 
      (match !r with None -> 
        if List.mem m (ftvmonad m2) then
          if (match (mflatten m2) with MVar (j,_) -> false | _ -> true) then
            (            raise (Type_error ("Failed occurs check unifying "^s1^", "^s2)))
          else ()
        else
          (r := Some m2)
      | Some m1' -> unify_monad m1' m2)
  | _,MVar _ -> unify_monad m2 m1
  | MBase (mid1,tl1),MBase (mid2,tl2) -> 
      if mid1 = mid2 
      then List.iter2 unify tl1 tl2
      else raise (Type_error ("Incompatible monads "^mid1^" "^mid2 ))
and unify t1 t2 = 
  let st1 = string_of_typ t1 in
  let st2 = string_of_typ t2 in
  let msg = ("unifying "^st1^" with "^st2^"; ") in 
  match t1,t2 with
    TVar (i,r), TVar(i',_) when i = i' -> (* User-defined types *)()
  | TVar (i,r),_ ->
      (match !r with None -> 
        let ftv2 = ftv t2 in
          if (List.mem i ftv2) then 
          if (match (ecr t2) with TVar (j,_) -> false | _ -> true) 
          then raise (Type_error ("Failed occurs check "^msg))
          else ()
        else
          r := Some t2
      | Some t1' -> unify t1' t2)
  | _,TVar _ -> unify t2 t1
  | TBase tid1,TBase tid2 -> 
      if tid1 = tid2 then () 
      else raise (Type_error ("Incompatible base types "^tid1^" "^tid2))
  | TFun(t1',t1''), TFun(t2',t2'') -> 
  begin try
      unify t1' t2'; unify_m t1'' t2''
  with Type_error s -> 
	raise (Type_error (msg^s)) end 
  | TTup(tl), TTup(tl') ->
      List.iter2 unify tl tl'
  | TApp(t1, tl1), TApp(t2, tl2) -> 
      unify t1 t2; (* these should always be base types *)
      List.iter2 unify tl1 tl2
  | TMorph(m1',m1''), TMorph(m2',m2'') -> unify_m m1' m2'; unify_m m1'' m2''
  | _,_ -> raise (Type_error ("Failed to match types "^msg))
and unify_m (m1,t1) (m2,t2) = unify_monad m1 m2; unify t1 t2

let rec unify_monad_args m1 m2 = let s1 = string_of_monad m1 in
            let s2 = string_of_monad m2 in
  match mflatten m1,mflatten m2 with
 | MBase (mid1,tl1),MBase (mid2,tl2) -> 
      if mid1 = mid2 
      then List.fold_left2 try_unify [] tl1 tl2
      else raise (Type_error ("Incompatible monads "^mid1^" "^mid2 ))
 | _,_ -> raise (Type_error ("Non constant monad constructors "^s1^" "^s2 ))
and try_unify acc t1 t2 = 
  let st1 = string_of_typ t1 in
  let st2 = string_of_typ t2 in
  let msg = ("try_unifying "^st1^" with "^st2^"; ") in 
  match t1,t2 with
    TVar (i,r), TVar(i',_) when i = i' -> (* User-defined types *)acc
  | TVar (i,r),_ ->
      (match !r with None -> 
        let ftv2 = ftv t2 in
          if (List.mem i ftv2) then 
            if (match (ecr t2) with TVar (j,_) -> false | _ -> true) 
            then raise (Type_error ("Failed occurs check "^msg))
            else acc
          else ((i,t2),fun () -> r:= Some t2) ::acc
      | Some t1' -> try_unify acc t1' t2)
  | _,TVar _ -> try_unify acc t2 t1
  | TBase tid1,TBase tid2 -> 
      if tid1 = tid2 then acc 
      else raise (Type_error ("Incompatible base types "^tid1^" "^tid2))
  | TFun(t1',t1''), TFun(t2',t2'') -> 
  begin try
      (try_unify acc t1' t2') @ (try_unify_m t1'' t2'')
  with Type_error s -> 
	raise (Type_error (msg^s)) end 
  | TTup(tl), TTup(tl') ->
      List.fold_left2 try_unify [] tl tl'
  | TApp(t1, tl1), TApp(t2, tl2) -> 
      let acc' = try_unify acc t1 t2 in 
      List.fold_left2 try_unify acc' tl1 tl2
  | TMorph(m1',m1''), TMorph(m2',m2'') -> (try_unify_m m1' m2') @ (try_unify_m m1'' m2'') @ acc
  | _,_ -> raise (Type_error ("Failed to match types "^msg))
and try_unify_m (m1,t1) (m2,t2) = let acc' = unify_monad_args m1 m2 in try_unify acc' t1 t2


exception Eval_error

(*
(* standard, capture-avoiding substitution *)
let rec subst (x,v) e =
  match e with
    Base _ | Susp _ -> e
  | Var y -> if x = y then e else v
(*  | Cons (c,l) -> Cons(c,List.map subst (x,v) l) *)
  | Lam (y,e') -> 
      if y = x then e else Lam (y,(subst (x,v) e'))
  | Rec (y,e') -> 
      if y = x then e else Rec (y,(subst (x,v) e'))
  | App (e1,e2) -> App (subst (x,v) e1, subst (x,v) e2)
  | Pair (e1,e2) -> Pair (subst (x,v) e1, subst (x,v) e2)
  | Index (b,e) -> Index (b,subst (x,v) e)
  | Let (y,e1,e2) -> 
      let e2' = if y = x then e2 else subst (x,v) e2 in
      Let (y,(subst (x,v) e1),e2')
  | Cond(e1,e2,e3) -> Cond(subst (x,v) e1, subst (x,v) e2, subst (x,v) e3)
  | Annot(e,t) -> Annot(subst (x,v) e, t)
  | Bind (m,e1,y,t,e2) -> 
      let e2' = if y = x then e2 else subst (x,v) e2 in
      Bind(m,(subst (x,v) e1),y,t,e2')
  | Unit (m,e) ->
      Unit(m,(subst (x,v) e))
*)
(* NG commented out evaluation
(* big-step operational semantics, parameterized by delta environment
   d, which maps variables to terms *)
let rec eval d e = 
  match e with 
    Base _ | Var _ | Lam _ | Susp _ -> e
  | App (e1,e2) -> 
      (match eval d e1 with
        Var x -> (d x) (eval d e2)
      | Susp f -> let v = eval d e2 in f v
      | Lam (x,e1) -> 
          let v = eval d e2 in
          eval d (subst (x,v) e1)
      | _ -> raise Eval_error)
  | Pair (e1,e2) -> Pair (eval d e1,eval d e2)
  | Index (b,e1) ->
      (match eval d e1 with
        Pair (v1,v2) -> if b then v1 else v2
      | _ -> raise Eval_error)
  | Let (x,e1,e2) ->
      eval d (App (Lam (x,e2), e1))
  | Annot(e,t) -> eval d e
  | Cond(e1,e2,e3) ->
      match eval d e1 with
	| Base(_,v) ->
	    if bool_of_string v
	    then
	      eval d e2 
	    else 
	      eval d e3
        | _ -> raise Eval_error
        

let default_eval_env x =
  match x with
    "plus" -> 
      (function y -> Susp (function z -> 
        match (y,z) with
          (Base ("int",i),(Base ("int",j))) -> 
            let r = (int_of_string i) + (int_of_string j) in
            Base("int",string_of_int r)
        | (Base ("float",f),(Base ("float",g))) -> 
            let r = (float_of_string f) +. (float_of_string g) in
            Base("float",(string_of_float r))
        | (Base ("bool",b),(Base ("bool",c))) -> 
            let r = (bool_of_string b) or (bool_of_string c) in
            Base("bool", (string_of_bool r))
        | _ -> raise Eval_error))
  | "eq" ->
      (function y -> Susp (function z ->
     match (y,z) with
       (Base (t,i),(Base (u,j))) ->
            if (t <> u) then raise Eval_error
            else Base("bool",string_of_bool (i = j))
     | _ -> raise Eval_error))
  | _ -> raise Eval_error
*)
(***********************************************************************)
(* Useful terms *)


(**** Example from Solving.tex ****)

(*
let idf = Lam("x",Var("x"))
let f = 
\x. let g = \y. eq x y in 
g 0, g (complex 1 i), g "hello" in 
f 0, f 0.0, f "bye" 

let ex1_e3 = Pair(App(Var "f",Base("int","0")),
                  Pair(App(Var "f",Base( "float","0.0")),
                       App(Var "f",Base( "string","bye"))))
let ex1_e2 = Pair(App(Var "g",Base("int","0")),
                  Pair(App(Var "g",Base( "complex","1 i")),
                       App(Var "g",Base( "string","hello"))))
let ex1_e1 = Lam("x",Let("g",Lam("y",App(App(Var "eq",Var "x"),Var "y")),ex1_e2))
let ex1 = Let("f",ex1_e1,ex1_e3)

*)
