(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)
module L = Lambd
open Util
open Typaux

let strict = ref true
let dot = ref false
(* Nodes of the graph are monads, edges are labels for morphisms*)

module Label = struct
  type t = L.ident
  let compare = compare
  let default = ""
end
(* Do everything on string representations.  This should dereference
     TVars properly, but it's slow and might give unexpected results if we
     unify and stuff is already in the graph. *)

module Monad = struct
  type t = L.monad
  let compare m1 m2 = compare (L.string_of_monad m1) (L.string_of_monad m2)
  let equal m1 m2 = (L.string_of_monad m1) = (L.string_of_monad m2)
  let hash m = Hashtbl.hash (L.string_of_monad m)
end

(* Membership based on monad equality *)
let rec monad_mem x  = function
    [] -> false
  | hd::tl -> if Monad.equal hd x then true else monad_mem x tl
let rec monad_assoc x  = function
    [] -> raise Not_found
  | (hd,v)::tl -> if Monad.equal hd x then v else monad_assoc x tl

module G = Graph.Imperative.Digraph.ConcreteLabeled(Monad)(Label)

(* output constraint graph as a DOT file *)
module Display = struct
  include G
  let vertex_name v = "\"" ^ (L.string_of_monad (V.label v)) ^ "\""
  let graph_attributes _ = [`Size (8.0,8.0)] (* size in inches *)
  let default_vertex_attributes _ = []
  let vertex_attributes _ = []
  let default_edge_attributes _ = []
  let edge_attributes e = let s = E.label e in [`Label s]
  let get_subgraph _ = None
end
module DotOutput = Graph.Graphviz.Dot(Display)

(* dump constraint graph g to file f *)
let dump_graph g f =  
  let oc = open_out f in
  DotOutput.output_graph oc g;
  close_out oc

open G

(* Fold on the binds with outer-edge v*)
let foldFlowsTo g f v acc = 
  let all = G.pred_e g v in
  G.fold_pred_e (fun e2 acc ->
    let m2,bvar = G.E.src e2, G.E.label e2 in
    if bvar.[0] = '*' then acc else
    try let e1 = List.find (fun e -> (G.E.label e) = "*"^bvar) all in
        f acc (bvar,e1,e2)
    with Not_found -> failwith "pred: Ill-formed dependence"
    ) g v acc

(* Fold on the binds that contain v as in-edge *)
let foldFlowsFrom g f v acc = 
  G.fold_succ_e (fun e2 acc ->     
    let m2,bvar = G.E.dst e2, G.E.label e2 in
    let all = G.pred_e g m2 in
    try 
    if bvar.[0] = '*' 
    then let label = String.sub bvar 1 (String.length bvar-1) in
      let e1 = List.find (fun e -> (G.E.label e) = label) all in
        f acc (label,e1,e2)
    else let label = "*"^bvar in
      let e1 = List.find (fun e -> (G.E.label e) = label) all in
        f acc (bvar,e1,e2)
    with Not_found -> failwith "succ:Ill-formed dependence"
    ) g v acc


(* make a constraint graph *)
let constraints_from_graph g = 
  G.fold_vertex (fun v l -> 
    let cs = foldFlowsTo g 
     (fun cs (bvar,e1,e2) -> (bvar,G.E.src e1, G.E.src e2) :: cs) v [] in    
    match cs with 
    | [] -> l
    | _ -> (List.rev cs,v)::l
  ) g []


let graph_from_constraints cs =
  let cslist = List.fold_left (fun l (c,m) -> 
   (List.fold_left (fun acc (l1,m1,m2) -> 
      (l1,m2,m) :: ("*"^l1,m1,m) (*:: ("**"^l1,m2,m1) :: (l1,m1,m)*) :: acc) [] c) @ l
  ) [] cs in
  let g = G.create () in
  List.iter (function (l,t1,t2) -> 
    let e = G.E.create t1 l t2 in
    G.add_edge_e g e) cslist;
  g

(*
let graph_from_constraints_cyc cs =
  let cslist = List.fold_left (fun l (c,m) -> 
   (List.fold_left (fun acc (l1,m1,m2) -> 
      (l1,m2,m) :: ("*"^l1,m1,m) (*:: ("**"^l1,m2,m1) :: (l1,m1,m)*) :: acc) [] c) @ l
  ) [] cs in
  let g = G.create () in
  List.iter (function (l,t1,t2) -> 
    let e = G.E.create t1 l t2 in
    G.add_edge_e g e) cslist;
  g
*)


let put_coer_term sln lbl tm =
  sln := (lbl,tm) :: !sln

let set_tvar_update_graph g tv typ = 
  (* Add the new type to the graph if not already present. *)
  G.add_vertex g typ;
  (* Move out-edges. *)
  G.iter_succ_e (fun e -> 
    G.remove_edge_e g e;
let e' = (G.E.create typ (G.E.label e) (G.E.dst e)) in
   G.add_edge_e g e';
		)
    g tv;
  (* Move in-edges. *)
  G.iter_pred_e (fun e -> 
    G.remove_edge_e g e;
    G.add_edge_e g (G.E.create (G.E.src e) (G.E.label e) typ)
		)
    g tv;
  (* Remove the type variable from the graph. *)
  G.remove_vertex g tv;
  (* Actually mutate the type variable. *)
  L.unify_monad tv typ


exception NoLub

let tsubmonad_all m sigma = 
  List.fold_left (fun m (k,nm) -> tsubmonad k nm m) m sigma

(* Refresh monad variables for a list of monads *)
let inst_monads ml =
  let fvars = union_all (List.map L.ftvmonad ml) in
  let subt = List.map (function tv -> (tv,L.freshtvar())) fvars in
  let refresh m = tsubmonad_all m subt in
  List.map refresh ml

let rec union_merge l = function
  | [] -> l
  | ((x,t),f) :: tl -> 
     let yes,no = List.partition (fun ((y,_),_) -> x=y) l in
     let yes',ff = List.fold_left (fun (acc,f'') ((y,t'),f') -> 
        let theta = L.try_unify [] t t' in
         (theta @ acc),
         (fun () -> f'();f''())) ([],fun () -> ()) yes in
     let res' = (((x,t),(fun () -> f(); ff())):: (union_merge no tl)) in
     yes' @ (List.map (fun ((y,ty),f') -> 
         (y,List.fold_left (fun t' (k,nt) -> tsub t' k nt ) ty (List.map fst yes')),f') 
         res')

module HMonad = struct
  type t = L.monad
  let equal m1 m2 = (*print_string "equal??";*) try 
    let nm1::[nm2] = inst_monads (m1::[m2]) in
    ignore (L.unify_monad_args nm1 nm2); print_endline "equal";true
    with L.Type_error e -> print_endline e; false 
  let hash m = Hashtbl.hash (L.string_of_monad m)
end

module HPairMonad = struct
  type t = L.monad * L.monad
  let equal (x1, y1) (x2, y2) = try  
    let nx1::ny1 :: nx2 :: [ny2] = (*inst_monads*) (x1::y1 :: x2::[y2]) in
    ignore
    (union_merge (L.unify_monad_args nx1 nx2) (L.unify_monad_args ny1 ny2)); true
    with L.Type_error _ -> false    
  let hash (x, y) = Hashtbl.hash (L.string_of_monad x, L.string_of_monad y)
end  

(*module MPairHashtbl = Hashtbl.Make(Graph.Util.HTProduct(Monad)(Monad))
module MHashtbl = Hashtbl.Make(Monad)
*)
module MPairHashtbl = Hashtbl.Make(HPairMonad)
module MHashtbl = Hashtbl.Make(HMonad) 

(* For each bind b:(M1,M2) < < M3, create an entry (M1,M2) -> (M3,b) *)
let bindset : (L.monad*string) MPairHashtbl.t = (MPairHashtbl.create 5)
let monadset: int MHashtbl.t = MHashtbl.create 5

(*
(* If M1 < M2, then create an entry (M1,M2) -> true *)
let order : bool MPairHashtbl.t = (MPairHashtbl.create 5)

let init_ordering (l:L.monad list list) =
  List.iter (fun l' -> ignore (List.fold_left (fun prevs hd ->  
                        if not (MPairHashtbl.mem order (hd,hd))
                        then MPairHashtbl.add order (hd,hd) true; 
    List.iter (fun x -> if not (MPairHashtbl.mem order (x,hd)) 
                        then MPairHashtbl.add order (x,hd) true; 
                        if not (MPairHashtbl.mem order (hd,x))
                        then MPairHashtbl.add order (hd,x) false) prevs; 
                        hd::prevs) [] l') 
    ) l 
*)
let rec init_bindset level = function
  | [] -> ()
  | Morph(n,b,m1,ps1,m2,ps2,m3,ps3) :: l when n<=level -> begin
    let mp1,mp2,mp3 = L.MBase(m1,ps1),L.MBase(m2,ps2),L.MBase(m3,ps3) in
(*print_endline ("add "^L.string_of_monad mp1^"?");
    if not (MHashtbl.mem monadset mp1) then MHashtbl.add monadset mp1 0;
  MHashtbl.iter (fun m _ -> print_endline ("& "^L.string_of_monad m)) monadset; 

print_endline ("add "^L.string_of_monad mp2^"?");
    if not (MHashtbl.mem monadset mp2) then MHashtbl.add monadset mp2 0;
  MHashtbl.iter (fun m _ -> print_endline ("& "^L.string_of_monad m)) monadset; 

print_endline ("add "^L.string_of_monad mp3^"?");
    if not (MHashtbl.mem monadset mp3) then MHashtbl.add monadset mp3 0;
  MHashtbl.iter (fun m _ -> print_endline ("& "^L.string_of_monad m)) monadset; 

    try let mold = MPairHashtbl.find_all bindset (mp1, mp2) in
      let bold = try monad_assoc mp3 mold with Not_found -> b in 
       if (bold <> b) then begin
        let st1 = L.string_of_monad mp1 in 
        let st2 = L.string_of_monad mp2 in 
        let st3 = L.string_of_monad mp3 in 
        Util.warning ("Multiple binds between "^st1^","^st2^" and "^st3^"\n")
        end
      else*) MPairHashtbl.add bindset (mp1,mp2) (mp3,b) (*;
      init_bindset level l
    with Not_found -> 
      MPairHashtbl.add bindset (mp1,mp2) (mp3,b);
      init_bindset level l *)
    end 
  | _ :: l -> init_bindset level l

(* TODO
 * Find all polymonand constructors M
 * For each M, check that (M,bot) > M and (bot,M) > M are defined.
 * Check existence for associativity.
 * Check paired morphisms.
 * Check functional closure.
 * Generate tests for Left, Right Id, and assoc laws equalities
 * Generate tests to ensure that binds with the same types have the same
 * semantics *)
let check_wf_morphisms s level = 
() (*
  init_bindset level s;
  MHashtbl.iter (fun m _ -> print_endline ("& "^L.string_of_monad m)) monadset; 
  let exist_all m1 m2 =
    try (MPairHashtbl.find_all bindset (m1,m2))
    with Not_found -> [] in
  let rec exists m1 m2 m3 = 
    List.exists (fun (m,_) -> Monad.equal m m) (exist_all m1 m2) in
  (* For each M, check that (M,bot) > M and (bot,M) > M are defined. *)
  MHashtbl.iter (fun m _ -> 
    if (exist_all m L.mBot) = [] 
    then warning ("Map missing for "^(L.string_of_monad m));
    if (exist_all L.mBot m) = []
    then warning ("Apply missing for "^(L.string_of_monad m))
  ) monadset; 
  (* Check existence for associativity. *)
  MPairHashtbl.iter (fun (m1,m2) (m12,_) ->
    let s1 = L.string_of_monad m1 in
    let s2 = L.string_of_monad m2 in 
       MHashtbl.iter (fun m3 _ -> 
       let s3 = L.string_of_monad m3 in
        List.iter (fun (m123,_) -> 
              if not (List.exists (fun (m23,_) -> exists m1 m23 m123) (exist_all m2 m3))
                then warning ("Assoc check failed for "^s1^","^s2^","^s3) 
           ) (exist_all m12 m3)
         ) monadset
    ) bindset 
*)  
(*
  let g = G.create () in
  List.iter (function (l,t1,t2) ->
    if G.mem_edge g t1 t2
    then
      let st1 = L.string_of_monad t1 in 
      let st2 = L.string_of_monad t2 in 
      failwith ("Multiple morphisms between "^st1^" and "^st2^"\n")
    else
    let e = G.E.create t1 l t2 in
    G.add_edge_e g e) cslist;
 G.clear g 
*)


(*
let reachable h m1 m2 =
 let rec reach h m1 m2 = 
 if Monad.equal m1 m2 then true, [(m1,m2,true)] else 
  try MPairHashtbl.find h (m1,m2),[]
  with Not_found -> 
  let hcp = MPairHashtbl.copy h in
  MPairHashtbl.add hcp (m1,m2) true;
  let findbinds m =
    MPairHashtbl.fold (fun (ma,mb) (mc,_) l -> 
      if ma = m then (mb,mc) :: l else l) bindset [] in
  let l1 = findbinds m1 in
  let l2 = findbinds m2 in
 let s1 = L.string_of_monad m1 in
let s2 = L.string_of_monad m2 in
  let exists = (fun mb mc (res,l) (mb1,mc1) -> 
    if res then (true,l) else
    let b1,acc1 = reach hcp mb1 mb in
    let b2,acc2 = reach hcp mc1 mc in
(*if b1 && b2 then Printf.printf "%s+%s<<%s, %s +%s<<%s\n" s2 (L.string_of_monad mb) (L.string_of_monad mc) s1 (L.string_of_monad mb1) (L.string_of_monad mc1);
*)    b1 && b2, (mb1,mb,b1) :: (mc1,mc,b2) :: acc1@acc2@l) in
  let forall (res,l) (mb,mc) = 
    if not res then (false,[]) else
    List.fold_left (exists mb mc) (false,[]) l1 in
  let res,acc = List.fold_left forall (true,[]) l2 in    
  MPairHashtbl.clear hcp;
  res,acc in
  let res,acc = (reach h m1 m2) in
  List.iter (fun (mb1,mc1,r) -> 
    try let r' = MPairHashtbl.find h (mb1,mc1) in assert (r=r')
    with Not_found -> MPairHashtbl.add h (mb1,mc1) r) 
      ((m1,m2,res)(*::(m2,m1,not res)*)::acc);
  res 
*)

(*
(* TODO The least monad, according to the monad order based on binds *)
let lub = function
 | [] -> raise NoLub
 | [hd] -> hd 
 | a1 :: tl -> List.fold_left 
       (fun ((m',v') as a) (m,v) -> try if MPairHashtbl.find order (m,m') 
        then (m,v) else a 
        with Not_found -> 
          print_endline (("Warning: incomparable "^L.string_of_monad m ^" and "^L.string_of_monad m')); a
 ) a1 tl
*)

(* Find a matching bind definition among all available binds; 
   returns a list of possible answers: the resulting monad, the bind name, and
   the updated names of the m1,m2 monads *) 
let find_bind h (m1,m2) = 
  let find (m,m') (mres,lab) acc = 
    let nm::nm'::[nmres] = inst_monads (m::m'::[mres]) in
    let m1'::[m2'] = (*inst_monads*) (m1::[m2]) in
    let ftvs = (L.ftvmonad m1') @ (L.ftvmonad m2') in
    try let sigma = union_merge (L.unify_monad_args nm m1') (L.unify_monad_args nm' m2') in 
        let sigma' = (*List.map (fun (((x,_),_) as acc)-> if List.mem x ftvs then acc)*) sigma in
        let acc' = (tsubmonad_all nmres (List.map fst sigma'),(lab,sigma')) :: acc in
acc'
    with L.Type_error _ -> acc in 
  MPairHashtbl.fold find h []


(* Find a particular bind definition (m1,m2) < m3 in the bindset *)
let find_a_bind h (m1,m2,m3) : string = 
  let bds = (find_bind bindset (m1, m2)) in
  (*let binds = List.map (fun (m,v) -> m,[bvar,v]) bds in *)
  let _,(bvar,_) = List.find (fun (m,v) -> L.string_of_monad m = L.string_of_monad m1) bds
  in bvar

let is_bot m = L.string_of_monad m == L.bottom
let is_const m = match L.mflatten m with L.MVar _ -> false | _ -> true 


(* Apply meta-simplifications *)
let meta_simplify g sln fmv t = 
    match L.mflatten t with
     L.MVar(mvar,_) as t when not (List.mem mvar fmv) -> begin
        let ms = L.string_of_monad t in 
        if not !quiet then print_endline ("Solving "^ms);

     begin 
     let preds = foldFlowsTo g (fun acc p -> p::acc) t [] in
     let succs = foldFlowsFrom g (fun acc p -> p::acc) t [] in
     match preds, succs with 
 (*    | [],_ -> (* No lower bounds *) 
        if not !quiet then print_endline ("No lower bound: Leaving "^ms^"unsolved")
     | _,[] -> (* No upper bound *)
        if not !quiet then print_endline ("No upper bound: Leaving "^ms^"unsolved")
*)
     (* METArule: (bot,m1) < < m and (m,m2) < m3 ==> m1/m *)
     | [(bvar,p1,p2)],_ when is_bot (G.E.src p1) -> 
       let sol = G.E.src p2 in
       if not !quiet then print_endline ("METAsolveUP&DROP:"^ms ^ " = "^L.string_of_monad sol);
       put_coer_term sln bvar (L.Var L.botBind);
       G.remove_edge_e g p1;
       G.remove_edge_e g p2;
       set_tvar_update_graph g t sol

     (* METArule: (m1,bot) < m and (m,m2) < m3 ==> m1/m *)
     | [(bvar,p1,p2)],_ when is_bot (G.E.src p2) -> 
       let sol = G.E.src p1 in
       if not !quiet then print_endline ("METAsolveUP:"^ms ^ " = "^L.string_of_monad sol);
       if is_const sol then
       begin try
         let coer = find_a_bind bindset (G.E.src p1, G.E.src p2,sol) in
         put_coer_term sln bvar (L.Var coer);
         G.remove_edge_e g p1;
         G.remove_edge_e g p2
       with Not_found -> print_endline ("Could not find bind "
           ^(L.string_of_constr ([(bvar,G.E.src p1,G.E.src p2)],sol)))
       end;
       set_tvar_update_graph g t sol

     (* METArule: (m1,m2) < m and (bot,m) < m3  ==> m3/m *)
     | _,[(bvar,p1,p2)] when is_bot (G.E.src p1) ->
       let sol = G.E.dst p1 in  
       if not !quiet then print_endline ("METAsolveDOWN&DROP:"^ms ^ " = "^L.string_of_monad sol );
       put_coer_term sln bvar (L.Var L.botBind);
       G.remove_edge_e g p1;
       G.remove_edge_e g p2;
       set_tvar_update_graph g t sol
     (* METArule: (m1,m2) < m and (m,bot) < m3 ==> m3/m *)
     | _,[(_,p1,p2)] when is_bot (G.E.src p2) ->
       let sol = G.E.dst p1 in  
       if not !quiet then print_endline ("METAsolveDOWN:"^ms ^ " = "^L.string_of_monad sol);
       set_tvar_update_graph g t sol

     | _ -> () 
     end end
   | _ -> ()

let meta_simplify_all g sln fmv = 
  let rec meta = function
    | [] -> ()
    | t :: rest -> meta_simplify g sln fmv t;  
      meta rest in
  meta (G.fold_vertex (fun x acc -> x:: acc) g [])

(* Removing cycles *)
(* Assumption: there is a constant inside each cycle (recursive functions must
  be annotated. Returns the list of the constraints that end in a constant,
  removes them from the graph.
*)

module Gcc = Graph.Components.Make(G)
let remove_cycles g sln =
  let ccl = Gcc.scc_list g in
  List.fold_left (fun intodo -> function
    | [] -> intodo
    | hd::tl as cc -> 
      if tl = [] && not (monad_mem hd (G.pred g hd)) then intodo else begin
      if not !quiet then print_string "Removing cycle ";
      let todo' = List.filter (fun n -> if not !quiet then print_string ("="^(L.string_of_monad n));
          (match L.mflatten n with L.MBase _ -> true | _ -> false)) cc in
      if not !quiet then print_endline "";
      let todo = if Util.empty todo' then begin
        print_endline "An annotation is needed for each recursive function";
        [hd] end
      else todo' in
      List.fold_left (fun accc s -> (foldFlowsTo g (fun acc (bvar,p1,p2) -> 
        G.remove_edge_e g p1;
        G.remove_edge_e g p2;
        p1::(p2::acc) ) s []) @ accc) intodo todo end) [] ccl

module H = Hashtbl.Make(G.V)
 
let topo_fold f g acc =
    let degree = H.create 100 in
    let todo = Queue.create () in
    let push x =
      H.remove degree x;
      Queue.push x todo
    in
     let rec walk acc =      

      if Queue.is_empty todo then
        let _ = H.iter (fun m v -> print_endline (L.string_of_monad m)) degree in
        if H.fold (fun m d n -> (match L.mflatten m with L.MBase _ -> n | _ -> n+1)) degree 0 = 0 
        then acc else failwith "cycle!"
(*	| Some(v, _) -> push v; walk acc
 *)     else
	let v = Queue.pop todo in
	let acc = f v acc in
	G.iter_succ_e
	  (fun p -> try
             let lab,dst = G.E.label p, G.E.dst p in
(* 
              if lab.[0] = '*' then () else
               let clab = String.sub lab 1 (String.length lab - 1) in
               let pp = List.find (fun x -> (G.E.label x) = clab) (G.succ_e g dst) in 
               G.E.dst pp
              else dst in
 *)            let d' = H.find degree dst in
	     if d' = 1 then push dst else H.replace degree dst (d'-1)
             with Not_found -> (* [x] already visited *)
	       ())
	  g v;
	walk acc
    in
    G.iter_vertex
      (fun v ->
	 let d = G.fold_pred_e (fun e n -> if (G.E.label e).[0]='*' then n else n+2) g v 0 in
	 if d = 0 then Queue.push v todo
	 else H.add degree v d)
      g;
    walk acc

let is_bot m = L.string_of_monad m == L.bottom
let is_const m = match L.mflatten m with L.MVar _ -> false | _ -> true 

let solve_binds fname g sln fmv =
  let l = List.rev (topo_fold (fun x l -> x ::l) g []) in
 
  let solve_step = ref 0 in
  List.iter (fun t ->  
    match L.mflatten t with
     L.MVar(mvar,_) as t when not (List.mem mvar fmv) -> begin
        let ms = L.string_of_monad t in 
        if not !quiet then print_endline ("Solving "^ms);
        if !dot && not !quiet then begin 
          solve_step := !solve_step + 1;
          dump_graph g (fname^"-"^ms^".dot") 
        end;
     (* Attempt meta-simplifications *)
    meta_simplify g sln fmv t;
     (* Normal solving *)
(* Compute the intersection of acceptable solutions for different lower bounds
 * A solution is a triple (m,(sigma,edges)) where 
 * m is the resulting monad,
 * sigma is the list of substitutions for bind names, and 
 * edges is the list of the edges between the involved lower bounds (to be removed) 
 *)
     let inter_sol s1 s2 =
       let find l (m',(sub',vs')) = 
         List.fold_left (fun acc (m,(sub,vs)) -> 
           try L.unify_monad m m'; 
               [m,(sub@sub',vs@vs')] 
           with L.Type_error _ -> []) [] l in
       List.concat (List.map (fun x -> find s2 x) s1) in
     
(* Compute a solution for each pair of lower bounds, then
 * then compute its intersection with the other solutions *)
     let bindsopt = foldFlowsTo g (fun acc (bvar,p1,p2) -> 
       let bds = (find_bind bindset (G.E.src p1, G.E.src p2)) in
       let binds = List.map (fun (m,v) -> m,([bvar,v],[p1;p2])) bds in
       match acc with 
      | None -> Some(binds) 
      | Some acc -> Some(inter_sol binds acc)) 
           t None in
     let binds = (match bindsopt with None -> [] | Some s -> s) in
     begin  try 
       let m3,(sub,preds) = try List.hd binds with _ -> raise NoLub in       
       List.iter (fun p -> G.remove_edge_e g p) preds;
       List.iter (fun (bvar,(bname,fsigma)) -> 
           (*List.iter (fun (x,t) -> print_string (" "^x^"/"^L.string_of_typ t)) (List.map fst fsigma); 
           print_newline (); *) 
           List.iter (fun xf -> xf ()) (List.map snd fsigma);
           put_coer_term sln bvar (L.Var bname)) sub;
       if not !quiet then print_endline (ms ^ " = "^L.string_of_monad m3);
       set_tvar_update_graph g t m3;

     with 
     | Not_found -> if not !quiet then print_endline ("No bind for "^ms)
     | NoLub -> if not !quiet then print_endline ("Leaving "^ms^"unsolved")
     end
  end
 | _ -> ()) l

(* Remove satisfiable constraints on constants *)
let solve_trans g sln =
  G.iter_vertex (fun v -> 
    foldFlowsTo g (fun acc (bvar,p1,p2) -> 
       let bds = (find_bind bindset (G.E.src p1, G.E.src p2)) in
       try let b,fsigma = monad_assoc (L.mflatten v) bds in
         List.iter (fun (_,f) -> f ()) fsigma;
         put_coer_term sln bvar (L.Var b); 
if not !quiet then print_endline ("Solving "^bvar^"="^b);
         G.remove_edge_e g p1;
         G.remove_edge_e g p2
       with Not_found -> ()
    ) v ()) g



(* solve ?w s cs fmv 
     solves the constraints cs using the initial signature s; 
     w is an optional string specifying the origin of the solving request,
       it may be used for debugging purpose; it should be a valid filename;
     fmv are the free variables that appear in the resulting type 
   Returns the unsolved constraints and the list of (var,term) morphism solutions *) 

let solve ?where (s:signature) (cs:L.constr list) (fmv:L.ident list) =
  List.iter (fun x -> print_string (x^", ")) fmv; print_newline ();
  let maxlevel = List.fold_left (fun m -> function Morph (x,_,_,_,_,_,_,_) -> max x m | _ -> m) 0 s in 
  let rec solve_level level cs = 
    if not !quiet then print_endline ("Solving at level "^string_of_int level); 
    let solv_cpt = ref 0 in
    if empty cs then [],[] else
    let pcs = Typaux.constraints_of_signature s in
    let fname = (Filename.chop_extension !infile)^ "-" ^ (match where with 
    | None -> solv_cpt := !solv_cpt + 1;
       "foo"^(string_of_int !solv_cpt)
    | Some s -> s) ^ "-" ^ (string_of_int level) in
    if not !quiet then print_endline ("Constraints for "^fname^":\n" ^ 
                                       (L.string_of_constrs "\n" cs)); 
    let sln = ref [] in
    let gr = graph_from_constraints cs in
    
    if !dot then dump_graph gr (fname ^"-initial.dot");
    let toadd = remove_cycles gr sln in
    if !dot then dump_graph gr (fname ^"-dag.dot");
    (*meta_simplify_all gr sln fmv;
  *)
    MPairHashtbl.clear bindset;  
    init_bindset level s;
    if (not !Util.debug)
    then solve_binds fname gr sln fmv
    else List.iter (meta_simplify gr sln fmv) (List.rev (topo_fold (fun x l -> x ::l) gr []));
    List.iter (fun x -> G.add_edge_e gr x) toadd;
    if !dot then dump_graph gr (fname^"-lub.dot");

    meta_simplify_all gr sln fmv;
    if !dot then dump_graph gr (fname ^"-meta.dot");
    solve_trans gr sln;
    if not !quiet then begin
      print_endline ("Solution:");
      List.iter (fun (l,tm) ->
          print_string ("  " ^ l ^ " := " ^ L.string_of_term tm ^ "\n"))
              (* In the current implementation, new choices are prepended.
                 Reverse to show the choices in the order they were made. *)
              (List.rev !sln)
    end;
     dump_graph gr (fname^"-uns.dot");  
     let unsolved = (constraints_from_graph gr) in
     if Util.empty unsolved || level >= maxlevel
     then (unsolved, !sln)
     else solve_level (level + 1) unsolved
  in
  let latex_flag = !Util.latex in
  Util.latex := false;
  let r = solve_level 0 cs in
  Util.latex := latex_flag;
  r
