(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)

module L=Lambd
open L
type level = int

exception NYI of string

type typ_env = (var * scheme) list 
type decl =
  | Monad of L.ident * L.ident * L.ident * typ list 
    (* type constructor,bind,unit, and default monad args *)
  | Morph of level * label * L.ident * typ list * L.ident * typ list * L.ident * typ list 
  | TyDecl of var * tvar list * (L.ident * typ list) list
type signature = decl list
type coer_env = constr list (* morphisms *)

let lookup_ty s tname : int (* returns type arity or raises Not_found *)= 
  match (try
    List.find (function TyDecl (v,l,_) when v=tname -> true | _ -> false) s
    with Not_found -> failwith ("Error: unbound type "^tname))
  with TyDecl(_,l,_) -> List.length l
  | _ -> failwith "by construction"

let lookup_cons s cname: (tvar list * L.ident * typ list) =
  let aux = function
  | TyDecl (tname,tvars,tdefs) -> 
    (try let args = List.assoc cname tdefs in Some (tvars,tname,args)
		with Not_found -> None)
  | _ -> None
  in let rec aux_rec = function
  | [] -> failwith ("Error: unbound type "^cname)
  | x :: l -> (match aux x with Some res->res | None -> aux_rec l)
  in aux_rec s


let lookup_monad (s:signature) (mname:string) : (var * var * typ list) = 
  let aux = function
  | Monad (m',_,_,_) -> mname=m'
  | _ -> false
 in match List.find aux s with Monad(_,fb,fu,args) -> (fb,fu,args) | _ -> failwith "Impossible" 

let mcons_vars_var s = (s^"_bind"), (s^"_unit")

(* Monad constructor variables *)
let id_of_monad m = match mflatten m with
  | MVar (s,r) -> s
  | MBase (s,tl) -> s^(String.concat "" (List.map (fun x -> String.capitalize (string_of_typ x)) tl)) 

let mcons_bind (m1,m2,m3) =
  "bind_"^id_of_monad m1^"_"^id_of_monad m2^"_"^id_of_monad m3
let mcons_unit m s = 
  match mflatten m with
  | MVar (v, _) -> "unit_"^v
  | m' -> "unit_"^(id_of_monad m')(*let (b,u,_) = lookup_monad s m' in u)*)

(*****************)
(* UGLY PRINTING *)
(*****************)

let string_of_signature sep ds =
  let rec string_of_decl = function
    | Monad (m,fb,fu,_) -> "(" ^ m ^ "," ^ fb ^ "," ^ fu ^ ")" 
    | Morph (_,lab,m1,tl1,m2,tl2,m3,tl3) -> lab ^ ":" ^  (string_of_monad (MBase(m1,tl1)))
          ^" + "^ (string_of_monad (MBase(m2,tl2)))
          ^" <| "^ (string_of_monad (MBase(m3,tl3)))
    | TyDecl (t,vl,dl) -> "type "^t^String.concat " " vl
  in
  String.concat sep (List.map string_of_decl ds)

let constraints_of_signature s = 
  List.fold_right (fun x l-> match x with 
   | Morph(_,lab,m1,tl1,m2,tl2,m3,tl3) -> (lab,MBase (m1,tl1),MBase (m2,tl2),MBase (m3,tl3)) :: l
   | Monad(m,_,fu,default)  -> (fu,mBot, mBot,MBase(m,default)) :: l
   | _ -> l) s [] 


(* substitute m for tvar k in type t *)
let rec tsub t k m =
  let t' = ecr t in
  match t' with
    TVar (i,r) -> (* r := None *)
      if (!r <> None) then failwith "oops!";
      if k = i then m else t'
  | TBase _ -> t'
  | TMorph (t1,t2) -> TMorph (tsubm t1 k m, tsubm t2 k m) 
  | TFun (t1,t2) -> TFun (tsub t1 k m, tsubm t2 k m)
  | TTup (tl) -> TTup (List.map (fun x -> tsub x k m) tl)
  | TApp (t, tl) -> TApp (tsub t k m, List.map (fun s -> tsub s k m) tl)
and tsubm (m,t) k nt = (tsubmonad k nt m,tsub t k nt) 
and tsubmonad k nt m = match mflatten m with
  | (MBase (m,ts)) as m'-> MBase(m,List.map (fun s -> tsub s k nt) ts)
  |  t -> t
  
(************************************************************************)
(* Coco Pretty printing *)

open Format;;


let ident = print_string;;
let kwd = print_string;;

let rec print_exp0 s = function
  | Base (t,v) -> 
      if (t = "int" || t = "bool" || t="float") then ident v
      else if t = "unit" then ident "()"
      else if (t = "string") then ident ("\"" ^ v ^ "\"")
      else if (t = "char") then ident ("\'" ^ v ^ "\'")
      else (kwd "{"; ident t; kwd ":"; ident v; kwd "}")
(*  | Cons (v,l) -> v ^ (match l with [] -> "" 
	| [x::l2] -> "("^string_of_term x ^ (List.map string_of_term l2)^ ")") *) 
  | Var s ->  ident s
  | Susp _ -> kwd "<susp>"
  | lam -> open_hovbox 1; kwd "("; print_lambda s lam; kwd ")"; close_box ()

and print_app s = function
  | e -> open_hovbox 2; print_other_applications s e; close_box ()

and print_annot s = function
  | Annot (e,t) -> 
      print_app s e; print_space (); kwd ":";       
      print_string (match t with 
          TLet t -> string_of_scheme t 
        | TBind t -> string_of_mtyp t)
  | e -> print_app s e

and print_other_applications s f =
  let (sg,mu) = s in
  match f with
  | EvApp (f, m,l) -> 
    ident f;   
    List.iter (fun x -> let fu = mcons_unit x sg in
          print_space (); ident fu) m; 
    List.iter (fun x -> print_space (); print_exp0 s x) l
  | EvAbs (ml,l, e) -> 

    let e' = List.fold_right (fun x ne ->
        let _,fu = mcons_vars_var x in
        Lam (fu,None,ne)) ml e in
    let e'' = List.fold_right (fun (x,t) ne -> Lam (x,t,ne)) l e' in
    print_lambda (sg,ml@mu) e''

  | App (f, arg) -> print_app s f; print_space (); print_exp0 s arg
  | Annot (e, t) -> print_exp0 s e
  | Let (y,Rec(z,e1),e2) when y=z -> 
      kwd ("let rec "^y^" ="); print_space();
      print_annot s e1; kwd " in"; print_space();
      print_lambda s e2; print_space()
  | Let (y,e1,e2) -> kwd ("let "^y^" ="); print_space();
      print_annot s e1; kwd " in"; print_space();
      print_lambda s e2; print_space()
  | Seq (e1,e2) -> 
      print_annot s e1; kwd ";"; print_space();
      print_lambda s e2; print_space()
  | Cond (e1,e2,e3) -> 
      open_hovbox 1; kwd "if"; print_space();
      print_exp0 s e1; print_space(); kwd "then"; print_space (); close_box ();
      print_app s e2; print_space (); kwd "else"; print_space ();
      print_app s e3
  | Bind (v,m,e1,e2) ->
                        let fb = (*if !Util.debug then mcons_bind m else*) (string_of_term v) in
                        ident fb; print_space (); print_exp0 s e1; 
                        print_space (); 
                        open_hovbox 1;
                        begin match e2 with 
                        | Lam _ -> kwd "("; print_lambda s e2; kwd ")"
                        | _ -> print_lambda s e2 end 
  | Unit (m,e) -> 
    let (sg,mu) = s in
    let fu = mcons_unit m sg in
    print_string fu; print_space(); print_exp0 s e 
  | f -> print_exp0 s f

and print_lambda sg = function
  | Lam (s, ty, lam) ->
      open_hovbox 1;
      kwd "fun "; 
      (match !Util.quiet,ty with 
       | true,Some t -> kwd "("; ident s; kwd ":"; ident (string_of_typ t); kwd ")"
       | _ -> ident s);
      print_space(); kwd "->"; print_space(); print_lambda sg lam;
      close_box()
  | Rec (s, lam) -> print_app sg (Let(s,Rec(s,lam),Var s))
  | Tuple (e::el) -> open_hovbox 1; kwd "<"; 
      print_lambda sg e; 
      List.iter (fun e2 -> kwd ","; print_space();  print_lambda sg e2) el;
      kwd ">"; close_box ()

  | Pair (e1,e2) -> open_hovbox 1; kwd "<"; print_lambda sg e1; 
      kwd ","; print_space();  print_lambda sg e2; kwd ">"; close_box ()
  | Index (b,e) -> open_hovbox 1; 
      if b then kwd "fst " else kwd "snd ";
      print_exp0 sg e; close_box()
  | e -> print_app sg e;;

let rec print_stritem sg = function
  | StrEmpty -> ()
  | StrTerm e -> print_lambda sg e
  | StrLet (b,x,e) -> 
    print_endline ("let "^(if b then "rec " else "")^x^" =");
    print_lambda sg e
  | StrCons (StrEmpty,s2) -> print_stritem sg s2
  | StrCons (s1,s2) -> print_stritem sg s1; print_endline ";;"; print_stritem sg s2

(* Flatten type constraints on polymorphic binds *)
(*let inst_bind init c (l,m1,ts1,m2,ts2,m3,ts3) =
  let tvars = union_all (List.map ftv (ts1@ts2@ts3) in
*)  


(****************************)
(* Conversion OCaml to Coco *)
(****************************)
open Camlp4.PreCast
module SQ = Syntax.Quotation
let _loc = Loc.ghost

module Caml =
  Camlp4.Printers.OCaml.Make
    (Camlp4OCamlParser.Make
      (Camlp4OCamlRevisedParser.Make
        (Camlp4.OCamlInitSyntax.Make(Ast)(Gram)(Quotation))));;


(* Naming convention for monad modules *)
let modname_of_typ = String.capitalize

(* TODO base name for parameterized monads? *)
let monad_base (m:monad) = 
  let name,t = match mflatten m with 
  | MVar (c,_) | MBase (c,[]) -> c,[]
  | MBase (c,cl) -> c,cl in
  modname_of_typ name,t

let rec ocaml_of_typ (sg:signature) = 
  function
    TBase i -> <:ctyp< $lid:i$ >> 
  | TVar (v,r) -> (try let _ = lookup_ty sg v in <:ctyp< $lid:v$>> with _ -> <:ctyp<'$v$>>)
  | TFun (t1,(m,t2)) ->
     let st1 = ocaml_of_typ sg t1 in
     let st2 = ocaml_of_typ sg t2 in
     let s = monad_base m in
     <:ctyp< $st1$ -> $monadtyp s st2$ >>
  | TMorph ((m1,t1),(m2,t2)) -> 
      let alpha = ocaml_of_typ sg t1 in
      let beta = ocaml_of_typ sg t2 in
      let base1 = monad_base m1 in
      let base2 = monad_base m2 in
      <:ctyp< $monadtyp base1 alpha$ -> $monadtyp base2 beta$>>
  | TApp (t, tl) -> 
      let t' = (ocaml_of_typ sg t) in
      let stl = List.map (ocaml_of_typ sg) tl in
      begin match stl with 
      | hd::tl ->
       List.fold_right (fun arg x -> <:ctyp< $x$  $arg$>>) stl t'
      | [] -> t' end
  | TTup [] -> <:ctyp< unit >> 
  | TTup [t] -> ocaml_of_typ sg t
  | TTup (hd::tl) -> let chd = ocaml_of_typ sg hd in
       let clist = List.fold_left 
          (fun x arg -> <:ctyp< $x$ * $ocaml_of_typ sg arg$>>) chd tl in
       <:ctyp< $tup:clist$ >>
and monadtyp (m,cl) st2 =
  if m = bottom then st2 
  else let x = (modname_of_typ m) in
       let fid = <:ident< $uid:x$ . $lid:"monad"$>> in
       let s = <:ctyp< $id:fid$>> in 
       if !Util.indexed then 
         let param = match cl with [] -> <:ctyp< unit >>
         | hd::tl -> List.fold_right 
           (fun arg x -> <:ctyp< $x$ * $ocaml_of_typ [] arg$>>) tl (ocaml_of_typ [] hd) in
         <:ctyp< $s$ $param$ $st2$>> 
       else begin
         assert (cl=[]);
         <:ctyp< $s$ $st2$>>
       end 

(* Terms reflecting the instantiation of type parameters for indexed monads *)
let rec term_of_typ s = function
   | TBase i -> if i = "int" then Base(i,"0")
     else if i = "float" then Base(i,".0")
     else if i = "bool" then Base(i,"true")
     else if i = "string" then Base (i,"0")
     else if i = "char" then Base(i,"0")
     else raise (NYI "term_of_typ:base")
   | TVar (i,_) -> term_of_typ s (TApp (TBase(i),[])) 
   | TTup (tl) -> Tuple (List.map (term_of_typ s) tl)
   | TApp (TBase(tname),tl) -> 
     begin match (try
      List.find (function TyDecl (v,_,_) when v=tname -> true | _ -> false) s
       with Not_found -> failwith ("Error: unbound type "^tname))
     with TyDecl(_,tvars,(cons,targs)::_) ->  
       let subst = List.combine tvars tl in
       let ntargs = List.fold_right (fun (k,m) res -> List.map (fun x -> tsub x k m) res) subst targs in
        App (Var cons, Tuple (List.map (term_of_typ s) ntargs))
     | _ -> failwith "by construction" end
   | t -> raise (NYI ("term_of_typ "^string_of_typ t))
   

 (* 
let ocamlbind brec x e =
  let p = <:patt< $lid:x$>> in 
  let b = <:binding< $pat:p$ = $exp:e$>> in
  if brec then <:str_item< value rec $b$>>
  else <:str_item< value $b$>>

let rec ocaml_of_term =
  let cpt = ref 0 in
  fun e s -> 
  match e with
  | Base (t,v) -> 
      if t = "int" then <:expr< $int:v$>>
      else if t = "bool" then <:expr< $`bool:bool_of_string v$>>
      else if t = "float" then <:expr< $flo:v$>>
      else if t = "unit" then <:expr< $lid:"()"$>>
      else if (t = "string") then <:expr< $str:v$>>
      else if (t = "char") then <:expr< $chr:v$>>
      else raise (NYI "ocaml_of_term: Base")
(*  | Cons (v,l) -> v ^ (match l with [] -> "" 
	| [x::l2] -> "("^string_of_term x ^ (List.map string_of_term l2)^ ")") *) 
  | Var v ->  <:expr< $lid:v$>>
  | EvApp (f, ml,l) -> cpt:=!cpt+1;
      let fev t e = 
        let tb = ocaml_of_term t s in
        <:expr< $e$ $tb$ >> in
      let fres e = List.fold_right fev l e in

      let fid = <:ident< $lid:f$ >> in
      if Util.empty ml 
      then let eres = <:expr< $id:fid$ >> in 
           <:expr< $fres eres$ >>
      else
      let mname = modname_of_typ f in
      let instance = mname^(string_of_int !cpt) in
      let mmod = <:ident< $uid:mname$>> in
      let minstance = <:ident< $uid: instance$>> in
      let fargs m e = 
        let b,_ = monad_base m in
        let mb = <:ident< $uid:b$>> in
        let mmb = <:module_expr< $id:mb$ >> in
        <:module_expr< $e$ $mmb$ >> in
      let mbound = List.fold_right fargs ml <:module_expr< $id:mmod$>> in
      let res = <:ident< $id:minstance$ . $id:fid$ >> in
      let eres = <:expr< $id:res$ >> in
      <:expr< let module $instance$ = $mbound$ in $fres eres$>>   
  | App (f, Tuple []) -> ocaml_of_term f s
  | App (f, arg) -> 
     let sf = ocaml_of_term f s in
     let sarg = ocaml_of_term arg s in
     <:expr< $sf$ $sarg$>>
  | Let (y,EvAbs(xl,morphs,e1),e2)
  | Let (y,Annot(EvAbs(xl,morphs,e1),_),e2) ->
      let se1 = ocaml_of_term (L.lam morphs e1) s in
      let se2 = ocaml_of_term e2 s in
      if Util.empty morphs
      then <:expr< let $lid:y$ = $se1$ in $se2$ >> 
      else 
      let mname = modname_of_typ y in
      let mbound = evabs false y xl se1 s in
      <:expr< let module $mname$ = $mbound$ in $se2$>>
  | Let (y,Rec(z,Annot(EvAbs(xl,morphs,e1),_)),e2)
  | Let (y,Rec(z,EvAbs(xl,morphs,e1)),e2) when y=z ->
      let se1 = ocaml_of_term (L.lam morphs e1) s in
      let se2 = ocaml_of_term e2 s in
      if Util.empty morphs
      then <:expr< let rec $lid:y$ = $se1$ in $se2$ >> 
      else 
      let mname = modname_of_typ y in
      let mbound = evabs true y xl se1 s in
      <:expr< let module $mname$ = $mbound$ in $se2$>>

  | Let (y,Rec(z,e1),e2) when y=z -> 
      let se1 = ocaml_of_term e1 s in
      let se2 = ocaml_of_term e2 s in
      let p = <:patt< $lid:y$>> in 
      let binding = <:binding< $pat:p$ = $exp:se1$>> in 
      <:expr< let rec $binding$ in $se2$>>
  | Let (y,e1,e2) -> 
      let se1 = ocaml_of_term e1 s in
      let se2 = ocaml_of_term e2 s in
      let p = <:patt< $lid:y$>> in 
      let binding = <:binding< $pat:p$ = $exp:se1$>> in 
      <:expr< let $binding$ in $se2$>>
  | Seq (e1,e2) -> 
      let se1 = ocaml_of_term e1 s in
      let se2 = ocaml_of_term e2 s in
      <:expr< $se1$; $se2$>>
  | Cond (e1,e2,e3) -> 
      let se1 = ocaml_of_term e1 s in
      let se2 = ocaml_of_term e2 s in
      let se3 = ocaml_of_term e3 s in
      <:expr< if $se1$ then $se2$ else $se3$ >>
  | Bind (m,e1,e2) -> 
      let base,_ = monad_base m in
      let se1 = ocaml_of_term e1 s in
      begin match e2 with 
      | Lam(y,_,e2') -> 
        if base = bottom then 
          if is_tmpvar y 
          then ocaml_of_term (subst_term (y,e1) e2') s
          else ocaml_of_term (Let(y,e1,e2')) s
        else         
      ((*try 
         let fb,_ = lookup_monad sg base in
         ocaml_of_term (App(App(Var fb,e1),e2)) s 
      with Not_found ->  *) 
          let fb = <:ident< $uid:base$>> in
          let se2 = ocaml_of_term e2 s in
          let fid = <:ident< $id:fb$ . $lid:"bind"$>> in
          let f = <:expr< $id:fid$ >> in
          let app1 = <:expr< $f$ $se1$ >> in
          <:expr< $app1$ $se2$ >>)
        | _ -> failwith "lambda by construction" end
  | Unit (m,e) -> 
      let base,cl = monad_base m in 
      if base = bottom then ocaml_of_term e s
      else
      ((*try 
         let _,fu = lookup_monad sg base in
         ocaml_of_term (App(Var fu,e)) s 
      with Not_found -> 
      *)
      let fb = <:ident< $uid:base$>> in
      let fid = <:ident< $id:fb$ . $lid:"unit"$>> in
      let f = <:expr< $id:fid$ >> in
      let se1 = ocaml_of_term e s in
      if !Util.indexed then
       let to_expr arg = ocaml_of_term (term_of_typ s arg) s in 
       let params = match cl with
       | [] -> <:expr< () >> 
       | hd::tl -> List.fold_right 
         (fun arg x -> 
           <:expr< $x$, $to_expr arg$>>
         ) tl (to_expr hd) in
       let param = <:expr< $tup:params$ >> in
       <:expr< $f$ $param$ $se1$ >>
      else       
       <:expr< $f$ $se1$ >>)
  | Lam (y,t,e1) -> (* TODO compile the type constraint *)
      let se1 = ocaml_of_term e1 s in
      let p = <:patt< $lid:y$>> in
      let pt = match t with
        | Some (TMorph _ as t') -> 
           let st = ocaml_of_typ s t' in
           <:patt< ($p$ : $st$) >> 
        | _ -> p in
      let m = <:match_case< $pat:pt$ -> $se1$>> in
      <:expr< fun [$m$] >>
  | Rec (y,e1) -> (*TODO*) ocaml_of_term (Let(y,Rec(y,e1),Var y)) s
  | Tuple [] -> <:expr< () >>
  | Tuple (e::el) -> 
      let e' = List.fold_left (fun l x -> <:expr< $l$, $ocaml_of_term x s$>>)
      (ocaml_of_term e s) el in
      <:expr< $tup:e'$ >>
  | Pair (e1,e2) ->       
      let se1 = ocaml_of_term e1 s in
      let se2 = ocaml_of_term e2 s in
      let tuple = <:expr< $se1$,$se2$ >> in
      <:expr< ($tup:tuple$)>> 
  | Index (b,e1) -> 
      let se1 = ocaml_of_term e1 s in
      let f = if b then <:expr< $lid:"fst"$>> else <:expr< $lid:"snd"$>> in
      <:expr< $f$ $se1$ >>
  | Susp _ -> raise (NYI "Susp only in eval") 
  | Annot (e,t) -> ocaml_of_term e s  
  | EvAbs _ -> failwith "EvAbs only occurs under Let"

and evabs brec y xl se1 sg =
      let me = ocamlbind brec y se1 in
      let monadtyp = <:ident< $uid:"MONAD"$>> in
      let monadmodtyp = <:module_type< $id:monadtyp$ >> in
      let ffunctor x e = 
        let name = modname_of_typ x in
        <:module_expr< functor ($uid:name$ : $mtyp:monadmodtyp$) -> $mexp:e$>> in
      List.fold_right ffunctor xl <:module_expr< struct $me$ end >>
      


(* TODO collect all the instantiations of the morphism/monad, 
and define the corresponing modules before declaring the morphisms *)
let ocaml_of_decl (* monads: ident * typ list *)= function
  | Monad (m,fbind,funit,default) when m <> bottom -> 
    let mname = modname_of_typ m in
    let alpha = <:ctyp< 'a>> in
    let beta = <:ctyp< 'b>> in
    let gamma = <:ctyp< 'c>> in
    let d b= 
       let fid = <:ident< $uid:mname$ . $lid:"monad"$>> in
       let s = <:ctyp< $id:fid$>> in 
       if !Util.indexed
       then <:ctyp< $s$ $gamma$ $b$>>
       else <:ctyp< $s$ $b$>> in

    let mt = <:module_type< $uid:"MONAD"$ >> in
    let smod =         
        let tcons = <:ctyp< $lid:m$ >> in
        let consapp = <:ctyp< $tcons$ $alpha$ >> in
        let monad = <:ctyp< $lid:"monad"$ >> in
        let monadfield = if !Util.indexed 
          then <:ctyp< $monad$ $gamma$ $alpha$ >>
          else <:ctyp< $monad$ $alpha$ >> in
        let wcstr = <:with_constr< type $monadfield$ = $typ:consapp$>> in
        let mty = <:module_type< $mtyp:mt$ with $with_constr:wcstr$ >> in
        <:sig_item< module $uid:mname$ : $mtyp:mty$ >> in
    let bind1 = <:ctyp< $alpha$ -> $d beta$>> in
    let bind0 = <:ctyp< $bind1$ -> $d beta$>> in
    let tbind = <:ctyp< $d alpha$ -> $bind0$>> in
    let strbind = <:sig_item< value $lid:fbind$ : $typ:tbind$>> in
  (* do we need the type parameter here? 
    let tunit = <:ctyp< $gamma$ -> $alpha$ -> $d gamma alpha$>> in
  *)  let tunit = <:ctyp< $alpha$ -> $d alpha$>> in
    let strunit = <:sig_item< value $lid:funit$ : $typ:tunit$>> in
    <:sig_item< $smod$; $strbind$; $strunit$ >>
  | Monad (m,fbind,funit,default) -> <:sig_item< >> 
  | Morph (l,m1,t1,m2,t2) -> (* TODO handle parameters *)
    let alpha = <:ctyp< 'a>> in
    let d1 = monadtyp (m1,t1) in
    let d2 = monadtyp (m2,t2) in
    let t = <:ctyp< $d1 alpha$ -> $d2 alpha$>> in
    <:sig_item< value $lid:l$ : $typ:t$>>
  | TyDecl _ -> <:sig_item< >>
(*| TyDecl(tname,[],[]) -> <:sig_item< type $lid:tname$>>
  | TyDecl(tname,vars,[]) ->
    let cty = List.fold_left (fun x arg -> 
           let at = <:ctyp< '$arg$ >> in
           <:ctyp< $x$  $at$>>) <:ctyp< $lid:tname$ >> vars in
    <:sig_item< type $typ:cty$>>
  | TyDecl(tname,vars,hd::tl) ->
    let clist = List.map (fun x -> <:ctyp< '$lid:x$>>) vars in
    let ocaml_of_def (i,tl') = 
      match tl' with 
        [] -> <:ctyp< $uid:i$>>
      | hd::tl -> 
      let t = List.fold_right (fun t l ->
          let t' = ocaml_of_typ [] t in
          <:ctyp< $t'$ * l >>
        ) tl (ocaml_of_typ [] hd) in 
      <:ctyp< $uid:i$ of $t$>> in
    let def_aux arg d = 
      <:ctyp< $ocaml_of_def arg$ | $d$ >> in
    let cdefs = List.fold_right def_aux tl (ocaml_of_def hd) in
    let ty = Ast.TyDcl(_loc,tname,clist,cdefs,[]) in
    <:sig_item< type $ty$>>
*)

let print_mli file g s =
  let qlist = fun x l -> <:sig_item< $x$; $l$ >> in
    let alpha = <:ctyp< 'a>> in
    let beta = <:ctyp< 'b>> in
    let gamma = <:ctyp< 'c>> in
    let d b= if !Util.indexed 
      then <:ctyp< $lid:"monad"$ $gamma$ $b$>>
      else <:ctyp< $lid:"monad"$ $b$>> in
    let strmon = <:sig_item< type $d alpha$>> in
    let bind1 = <:ctyp< $alpha$ -> $d beta$>> in
    let bind0 = <:ctyp< $bind1$ -> $d beta$>> in
    let tbind = <:ctyp< $d alpha$ -> $bind0$>> in
    let strbind = <:sig_item< value $lid:"bind"$ : $typ:tbind$>> in
    let tunit = if !Util.indexed
      then <:ctyp< $gamma$ -> $alpha$ -> $d alpha$>>
      else <:ctyp< $alpha$ -> $d alpha$>> in
    let strunit = <:sig_item< value $lid:"unit"$ : $typ:tunit$>> in
    let strbindunit = <:sig_item< $strbind$; $strunit$>> in
    let si = <:sig_item< $strmon$; $strbindunit$ >> in 
  let m' = <:module_type< sig $sigi:si$ end >> in
  let mtype = <:sig_item< module type $uid:"MONAD"$ = $mtyp:m'$ >> in
  let idmt = <:module_type< $uid:"MONAD"$ >> in
  let wcstr = <:with_constr< type $typ:d alpha$ = $typ:alpha$>> in
  let idmty = <:module_type< $mtyp:idmt$ with $with_constr:wcstr$ >> in
  let idtype = <:sig_item< module $uid:modname_of_typ bottom$ : $mtyp:idmty$ >> in
  let idty = <:ctyp< $alpha$ -> $alpha$ >> in
  let idfuns = <:sig_item< value $lid:"id"$ : $typ: idty$>> in

  let ssig = (List.fold_right qlist (mtype::idtype::(List.map ocaml_of_decl s)) idfuns) in
  let fenv = fun (x,Scheme(vars,cs,t)) -> (* assert vars=cs=[] *)
      <:sig_item< value $lid:x$ : $typ:(ocaml_of_typ s t)$>> in
  let senv = match (List.map fenv g) with
    | [] -> ssig
    | hd::tl -> 
        let defs = List.fold_right qlist tl hd in   
        <:sig_item< $ssig$; $defs$ >> in
  let sopen = List.fold_left (fun acc fmli ->
    let dir,base = Util.split_filename fmli in (*TODO include dir into the module name *) 
    let lib = <:ident< $uid: modname_of_typ base$ >> in
    let sopen1 = <:sig_item< open $id:lib$ >> in
    <:sig_item< $acc$ ; $sopen1$ >>) <:sig_item<>> !Util.mli_input in
  Caml.print_interf ~output_file:file <:sig_item< $sopen$;$senv$>>

let rec ocaml_of_stritem s = function
  | StrEmpty -> <:str_item< >>
  | StrTerm e -> begin  <:str_item< $exp:ocaml_of_term e s$ >> end 
  | StrLet (b,y,EvAbs(xl,morphs,e1))
  | StrLet (b,y,Annot(EvAbs(xl,morphs,e1),_)) ->
      let se1 = ocaml_of_term (L.lam morphs e1) s in
      let pat = <:patt< $lid:y$ >> in
      if Util.empty morphs
      then 
        let bind = <:binding< $pat:pat$ = $se1$ >> in
        if b then <:str_item< value rec $bind$>>
        else <:str_item< value $bind$>>
      else 
        let mname = modname_of_typ y in
        let mbound = evabs b y xl se1 s in
        <:str_item< module $mname$ = $mbound$>>
 
  | StrLet (b,x,e) -> 
    let e' = ocaml_of_term e s in
    let pat = <:patt< $lid:x$ >> in
    let bind = <:binding< $pat:pat$ = $e'$ >> in
    if b then <:str_item< value rec $bind$>>
    else <:str_item< value $bind$>>
  | StrCons (s1,s2) -> 
    let s1' = ocaml_of_stritem s s1 in
    let s2' = ocaml_of_stritem s s2 in
    <:str_item< $s1'$; $s2'$ >>

let print_ml ?output_file e s =
  let res = ocaml_of_stritem s e in   
  match output_file with
  | None -> Caml.print_implem res
  | Some file -> 
    let _,base = Util.split_filename file in
    let out_mli = <:ident< $lid:"Lib"^base$ >> in
    let sopen = <:str_item< open $out_mli$ >> in 
    let sopenmli = List.fold_left (fun acc fmli ->
      let _,base2 = Util.split_filename fmli in
      let lib2 = <:ident< $uid: modname_of_typ base2$ >> in
      let sopen2 = <:str_item< open $id:lib2$ >> in
      <:str_item< $acc$ ; $sopen2$ >>) <:str_item<>> !Util.mli_input in
    Caml.print_implem ~output_file:file <:str_item< $sopen$;$sopenmli$;$res$ >>
*)
(****************************)
(* Conversion OCaml to Coco *)
(****************************)

(* Convert an Ocaml type into Coco type;
   id is the name of the binding where this type is used *)
let rec typ_of_ocaml id (sg:signature) = function
  | <:ctyp< $t1$ -> $t2$ >> -> 
    let m2,ct2 = match typapp_of_ocaml id sg t2 with
    | Pure t -> mBot, t
    | Monadic (m,t) -> m,t in
    begin match typapp_of_ocaml id sg t1 with
      | Pure ct1 -> TFun(ct1,(m2,ct2))
      | Monadic (m1,ct1) -> 
          let smorphism = "a custom morphism "^(string_of_monad m1) ^" << "^(string_of_monad m2)in
          print_endline ("Warning: function "^id^" seems to perform "^smorphism);
          TMorph((m1,ct1),(m2,ct2))
    end 
  | <:ctyp< $t1$ $t2$ >> as t -> 
    begin match typapp_of_ocaml id sg t with
    | Pure t -> t
    | Monadic t -> failwith (id^": unexpected monadic type")
    end
  | <:ctyp< $lid:i$ >> -> TBase i
  | <:ctyp< $tup:t$ >> -> 
       let rec aux_tup x l = 
         match x with 
         | <:ctyp< $x1$ * $x2$ >> -> 
           let l1 = aux_tup x1 l in
           aux_tup x2 l1
         | _ -> l@[typ_of_ocaml id sg x]
       in TTup(aux_tup t [])
  | <:ctyp< '$v$>> -> TVar (v,ref None)
  | _ -> raise (NYI "typ_of_ocaml")
and typapp_of_ocaml id (sg:signature) tapp= 
  let rec typapp_aux acc = function      
  | <:ctyp< $tcons$ $ta$>> ->
      let acc' = match ta with 
        | <:ctyp< >> -> acc
        | t -> (typ_of_ocaml id sg t) :: acc in
      typapp_aux acc' tcons  
  | t -> typ_of_ocaml id sg t, acc in 
  let tf,targs =  typapp_aux [] tapp in
  match targs with 
  | [] -> Pure tf
  | _ -> 
     let (content :: margs ) = List.rev targs in
     let tname = string_of_typ tf in
        begin try
          let _,_,nargs = lookup_monad sg tname in
          let declared = List.length nargs in
          let given = List.length margs in
          if declared = given then 
            Monadic (MBase(tname,List.rev margs),content)
	  else failwith (id^": Monad "^tname^" expects "^(string_of_int declared)
               ^" type parameters, but is here applied to "
               ^(string_of_int given)^" parameter(s)")
        with Not_found ->
          let nargs = lookup_ty sg tname in
          let given = List.length targs in
          if nargs = given then 
            Pure (TApp(TBase(tname),targs)) 
	  else failwith (id^": Type "^tname^" expects "^(string_of_int nargs)
               ^" parameters, but is here applied to "
               ^(string_of_int given)^" parameter(s)")

          end


let rec sig_of_ocaml (sg:signature) = function
  | <:sig_item< value $lid:id$ : $typ:ty$>> -> 
    let t = typ_of_ocaml id sg ty in
    [(id,Scheme(ftv t,[],t))],sg
  | <:sig_item< $s1$ ; $s2$>> ->
    let sg1,ts1 = sig_of_ocaml sg s1 in
    let sg2,ts2 = sig_of_ocaml ts1 s2 in
    sg1 @ sg2, ts2 
  | <:sig_item< type $typ:ty$ >> ->
    begin match ty with
    | Ast.TyDcl(_,tname,clist,cdefs,_) ->
      let params = List.map (function <:ctyp< '$lid:x$>> -> x 
	   | _ -> raise (NYI "sig_of_ocaml, type_decl, type param")) clist in
      let rec def_of = function
      | <:ctyp< >> -> []
      | <:ctyp< $uid:i$>> -> [i,[]] 
      | <:ctyp< $uid:i$ of $t$ >> -> 
        [i, (match typ_of_ocaml "" sg t with TTup tl -> tl | tl -> [tl])]
      | <:ctyp< $t1$ | $t2$ >> -> 
        (def_of t1) @ (def_of t2) 
      | _ -> [] (*raise (NYI "sig_of_ocaml, type_decl, def_of")*) in
     [],TyDecl(tname,params,def_of cdefs) :: sg
    | _ -> raise (NYI "sig_of_ocaml, type_decl")
    end    
  | <:sig_item< >> 
  | <:sig_item< open $_$ >> -> [],sg
  | _ -> raise (NYI "sig_of_ocaml")

let rec term_of_ocaml sg = function
  | <:expr< $int:v$>> -> Base ("int",v) 
  | <:expr< $flo:v$>> -> Base("float",v)
  | <:expr< ()>> -> Base("unit","()")
  | <:expr< $str:v$>> -> Base("string",v)
  | <:expr< $chr:v$>> -> Base("char",v)
  | <:expr< $lid:v$>> -> Var v
  | <:expr< $`bool:v$>> -> 
    begin match v with 
    | <:ident< $uid:i$ >> -> Base("bool",String.uncapitalize i)
    | <:ident< $lid:i$ >> -> Base("bool",i)
    | _ -> failwith "ocaml_of_term: decoding a boolean"
    end
  | <:expr< $e1$ $e2$ >> -> 
    let ce2 = term_of_ocaml sg e2 in 
    begin match e1 with
    | <:expr< $lid:s$ >> -> 
      if s = "fst" then Index(true,ce2) else
      if s = "snd" then Index(false,ce2) else
      App(Var s, ce2)
    | _ -> App(term_of_ocaml sg e1,ce2)
    end
  | <:expr< if $e1$ then $e2$ else $e3$ >> ->
      Cond(term_of_ocaml sg e1,term_of_ocaml sg e2,term_of_ocaml sg e3)
  | <:expr< let $b$ in $e$ >> -> 
      let ((x,_),ce1) = binding_of_ocaml sg b in
      let ce2 = term_of_ocaml sg e in
      Let(x,ce1,ce2)
  | <:expr< let rec $b$ in $e$ >> -> 
      let ((x,topt),ce1) = binding_of_ocaml sg b in
      let ce2 = term_of_ocaml sg e in
      begin match topt with 
      | None -> Let(x,Rec(x,ce1),ce2)
      | Some t -> Let(x,Rec(x,Annot(ce1,TLet(Scheme([],[],t)))),ce2) end
  | <:expr< fun [ $a$ ] >> -> let x,e = match_of_ocaml sg a in Lam(x,None,e)
  | <:expr< $e1$ , $e2$ >> -> Pair(term_of_ocaml sg e1,term_of_ocaml sg e2)      
  | <:expr< $tup:e$ >> -> 
       let rec aux_tup x l = 
         match x with 
         | <:expr< $x1$ , $x2$ >> -> 
           let l1 = aux_tup x1 l in
           aux_tup x1 l1
         | _ -> l@[term_of_ocaml sg x]
       in Tuple(aux_tup e [])
  | <:expr< $e1$ ; $e2$ >> -> Seq(term_of_ocaml sg e1,term_of_ocaml sg e2)      
  | <:expr< $seq:e$ >> -> term_of_ocaml sg e
  | <:expr< ($e$ : $t$) >> -> 
    let typannot = 
       match typapp_of_ocaml "annot_mon" sg t with 
       | Pure p -> TLet(Scheme([],[],typ_of_ocaml "annot_pure" sg t))
       | Monadic(m,p) -> TBind(m,p) in
    Annot(term_of_ocaml sg e,typannot)
  | _ -> raise (NYI "term_of_ocaml")

and binding_of_ocaml sg = function
  | <:binding< $pat:p$ = $exp:e$ >> -> 
    (pattern_of_ocaml sg p, term_of_ocaml sg e)
  | _ -> raise (NYI "binding_of_ocaml")
and pattern_of_ocaml sg = function
  | <:patt< _ >> -> "_",None
  | <:patt< () >> -> "()",None
  | <:patt< $lid:i$ >> -> i,None
  | <:patt< ($p$:$t$) >> -> let s,_ = pattern_of_ocaml sg p in 
    s, Some (typ_of_ocaml "patt" sg t)
  | _ -> raise (NYI "pattern_of_ocaml")
and match_of_ocaml sg = function
  | <:match_case< $p$ -> $e$ >> -> (fst (pattern_of_ocaml sg p), term_of_ocaml sg e)
  | _ -> raise (NYI "match_case_of_ocaml")
  
let parse_ocamli f =
  let ic = open_in f in
  let strm = Stream.of_channel ic in
  let res = Caml.parse_interf (Loc.mk f) strm in
  close_in ic; res

let parse_ocaml ic sg =
  let strm = Stream.of_channel ic in
  let res = Caml.parse_implem (Loc.mk "mlinput") strm in
  let rec aux = function 
  | <:str_item< $exp:expr$ >> -> StrTerm (term_of_ocaml sg expr)
  | <:str_item< $s1$;$s2$ >> -> StrCons (aux s1, aux s2)
  | <:str_item< >> -> StrEmpty
  | <:str_item< value $b$ >> -> 
    begin match b with 
    | <:binding< $p$ = $exp:e$ >> -> 
      begin match p with
      | <:patt< $lid:x$ >> -> StrLet(false,x,term_of_ocaml sg e)
      | _ -> failwith "expecting simple var top-level bindings" end 
    | _ -> failwith "expecting simple patt top-level bindings" end 
  | <:str_item< value rec $b$ >> -> 
    begin match b with 
    | <:binding< $p$ = $exp:e$ >> -> 
      begin match p with
      | <:patt< $lid:x$ >> -> StrLet(true,x,term_of_ocaml sg e)
      | _ -> failwith "expecting simple var top-level bindings" end 
    | _ -> failwith "expecting simple patt top-level bindings" end 
 | _ -> failwith "Expecting a top-level expression to convert"
  in aux res
